import math

import cupy
from cupy._core._scalar import get_typename
from cupy import _util
import cupyx.scipy.linalg
import cupyx.scipy.signal._signaltools
from cupyx.scipy.signal._arraytools import (
    axis_slice,
    axis_reverse,
    const_ext,
    even_ext,
    odd_ext,
)


def firfilter(b, x, axis=-1, zi=None):
    """
    Filter data along one-dimension with an FIR filter.

    Filter a data sequence, `x`, using a digital filter. This works for many
    fundamental data types (including Object type). Please note, cuSignal
    doesn't support IIR filters presently, and this implementation is optimized
    for large filtering operations (and inherently depends on fftconvolve)

    Parameters
    ----------
    b : array_like
        The numerator coefficient vector in a 1-D sequence.
    x : array_like
        An N-dimensional input array.
    axis : int, optional
        The axis of the input data array along which to apply the
        linear filter. The filter is applied to each subarray along
        this axis.  Default is -1.
    zi : array_like, optional
        Initial conditions for the filter delays.  It is a vector
        (or array of vectors for an N-dimensional input) of length
        ``max(len(a), len(b)) - 1``.  If `zi` is None or is not given then
        initial rest is assumed.  See `lfiltic` for more information.

    Returns
    -------
    y : array
        The output of the digital filter.
    zf : array, optional
        If `zi` is None, this is not returned, otherwise, `zf` holds the
        final filter delay values.
    """

    return cupyx.scipy.signal.lfilter(b, cupy.ones((1,)), x, axis, zi)


def firfilter_zi(b):
    """
    Construct initial conditions for lfilter for step response steady-state.
    Compute an initial state `zi` for the `lfilter` function that corresponds
    to the steady state of the step response.
    A typical use of this function is to set the initial state so that the
    output of the filter starts at the same value as the first element of
    the signal to be filtered.
    Parameters
    ----------
    b : array_like (1-D)
        The IIR filter coefficients. See `lfilter` for more
        information.
    Returns
    -------
    zi : 1-D ndarray
        The initial state for the filter.
    See Also
    --------
    lfilter, lfiltic, filtfilt
    Notes
    -----
    A linear filter with order m has a state space representation (A, B, C, D),
    for which the output y of the filter can be expressed as::
        z(n+1) = A*z(n) + B*x(n)
        y(n)   = C*z(n) + D*x(n)
    where z(n) is a vector of length m, A has shape (m, m), B has shape
    (m, 1), C has shape (1, m) and D has shape (1, 1) (assuming x(n) is
    a scalar).  lfilter_zi solves::
        zi = A*zi + B
    In other words, it finds the initial condition for which the response
    to an input of all ones is a constant.
    Given the filter coefficients `a` and `b`, the state space matrices
    for the transposed direct form II implementation of the linear filter,
    which is the implementation used by scipy.signal.lfilter, are::
        A = scipy.linalg.companion(a).T
        B = b[1:] - a[1:]*b[0]
    assuming `a[0]` is 1.0; if `a[0]` is not 1, `a` and `b` are first
    divided by a[0].
    """
    a = cupy.ones(1, dtype=cupy.float32)
    n = len(b)

    # Pad a with zeros so they are the same length.
    a = cupy.r_[a, cupy.zeros(n - len(a), dtype=a.dtype)]
    IminusA = (cupy.eye(n - 1, dtype=cupy.result_type(a, b))
               - cupyx.scipy.linalg.companion(a).T)
    B = b[1:] - a[1:] * b[0]
    # Solve zi = A*zi + B
    zi = cupy.linalg.solve(IminusA, B)

    return zi


def _validate_pad(padtype, padlen, x, axis, ntaps):
    """Helper to validate padding for filtfilt"""
    if padtype not in ["even", "odd", "constant", None]:
        raise ValueError(
            (
                "Unknown value '%s' given to padtype.  padtype "
                "must be 'even', 'odd', 'constant', or None."
            )
            % padtype
        )

    if padtype is None:
        padlen = 0

    if padlen is None:
        # Original padding; preserved for backwards compatibility.
        edge = ntaps * 3
    else:
        edge = padlen

    # x's 'axis' dimension must be bigger than edge.
    if x.shape[axis] <= edge:
        raise ValueError(
            "The length of the input vector x must be greater "
            "than padlen, which is %d." % edge
        )

    if padtype is not None and edge > 0:
        # Make an extension of length `edge` at each
        # end of the input array.
        if padtype == "even":
            ext = even_ext(x, edge, axis=axis)
        elif padtype == "odd":
            ext = odd_ext(x, edge, axis=axis)
        else:
            ext = const_ext(x, edge, axis=axis)
    else:
        ext = x
    return edge, ext


def firfilter2(
    b, x, axis=-1, padtype="odd", padlen=None, method="pad", irlen=None
):
    """
    Apply a digital filter forward and backward to a signal.
    This function applies a linear digital filter twice, once forward and
    once backwards.  The combined filter has zero phase and a filter order
    twice that of the original.
    The function provides options for handling the edges of the signal.
    The function `sosfiltfilt` (and filter design using ``output='sos'``)
    should be preferred over `filtfilt` for most filtering tasks, as
    second-order sections have fewer numerical problems.
    Parameters
    ----------
    b : (N,) array_like
        The numerator coefficient vector of the filter.
    x : array_like
        The array of data to be filtered.
    axis : int, optional
        The axis of `x` to which the filter is applied.
        Default is -1.
    padtype : str or None, optional
        Must be 'odd', 'even', 'constant', or None.  This determines the
        type of extension to use for the padded signal to which the filter
        is applied.  If `padtype` is None, no padding is used.  The default
        is 'odd'.
    padlen : int or None, optional
        The number of elements by which to extend `x` at both ends of
        `axis` before applying the filter.  This value must be less than
        ``x.shape[axis] - 1``.  ``padlen=0`` implies no padding.
        The default value is ``3 * max(len(a), len(b))``.
    method : str, optional
        Determines the method for handling the edges of the signal, either
        "pad" or "gust".  When `method` is "pad", the signal is padded; the
        type of padding is determined by `padtype` and `padlen`, and `irlen`
        is ignored.  When `method` is "gust", Gustafsson's method is used,
        and `padtype` and `padlen` are ignored.
    irlen : int or None, optional
        When `method` is "gust", `irlen` specifies the length of the
        impulse response of the filter.  If `irlen` is None, no part
        of the impulse response is ignored.  For a long signal, specifying
        `irlen` can significantly improve the performance of the filter.
    Returns
    -------
    y : ndarray
        The filtered output with the same shape as `x`.
    Notes
    -----
    When `method` is "pad", the function pads the data along the given axis
    in one of three ways: odd, even or constant.  The odd and even extensions
    have the corresponding symmetry about the end point of the data.  The
    constant extension extends the data with the values at the end points. On
    both the forward and backward passes, the initial condition of the
    filter is found by using `lfilter_zi` and scaling it by the end point of
    the extended data.
    When `method` is "gust", Gustafsson's method [1]_ is used.  Initial
    conditions are chosen for the forward and backward passes so that the
    forward-backward filter gives the same result as the backward-forward
    filter.
    The option to use Gustaffson's method was added in scipy version 0.16.0.
    References
    ----------
    .. [1] F. Gustaffson, "Determining the initial states in forward-backward
           filtering", Transactions on Signal Processing, Vol. 46, pp. 988-992,
           1996.
    """
    b = cupy.atleast_1d(b)
    x = cupy.asarray(x)
    if method not in ["pad", "gust"]:
        raise ValueError("method must be 'pad' or 'gust'.")

    if method == "gust":
        raise NotImplementedError("gust method not supported yet")

    # method == "pad"
    edge, ext = _validate_pad(padtype, padlen, x, axis, ntaps=len(b))

    # Get the steady state of the filter's step response.
    zi = firfilter_zi(b)

    # Reshape zi and create x0 so that zi*x0 broadcasts
    # to the correct value for the 'zi' keyword argument
    # to lfilter.
    zi_shape = [1] * x.ndim
    zi_shape[axis] = zi.size
    zi = cupy.reshape(zi, zi_shape)
    x0 = axis_slice(ext, stop=1, axis=axis)

    # Forward filter.
    (y, zf) = firfilter(b, ext, axis=axis, zi=zi * x0)

    # Backward filter.
    # Create y0 so zi*y0 broadcasts appropriately.
    y0 = axis_slice(y, start=-1, axis=axis)
    (y, zf) = firfilter(b, axis_reverse(y, axis=axis), axis=axis, zi=zi * y0)

    # Reverse y.
    y = axis_reverse(y, axis=axis)

    if edge > 0:
        # Slice the actual signal from the extended signal.
        y = axis_slice(y, start=edge, stop=-edge, axis=axis)

    return cupy.copy(y)


_freq_shift_kernel = cupy.ElementwiseKernel(
    "float64 x, float64 c", "complex128 out",
    "out = thrust::polar(x, c * i);",
    "_freq_shift_kernel",
)


def freq_shift(x, freq, fs):
    """
    Frequency shift signal by freq at fs sample rate

    Parameters
    ----------
    x : array_like, complex valued
        The data to be shifted.
    freq : float
        Shift by this many (Hz)
    fs : float
        Sampling rate of the signal
    domain : string
        freq or time
    """
    x = cupy.asarray(x)
    c = -2 * math.pi * freq / fs
    return _freq_shift_kernel(x, c)


_SUPPORTED_TYPES = [
    "float32_complex64",
    "complex64_complex64",
    "float64_complex128",
    "complex128_complex128",
]


_CHANNELIZER_KERNEL_PREAMBLE = r"""
#include <cooperative_groups.h>
#include <cooperative_groups/reduce.h>
#include <cupy/complex.cuh>
#include <cupy/cuda_workaround.h>

namespace std {
template< class T, class U >
inline constexpr bool is_same_v = is_same<T, U>::value;
}

namespace cg = cooperative_groups;
"""


_CHANNELIZER_8X8_KERNEL = _CHANNELIZER_KERNEL_PREAMBLE + r"""
///////////////////////////////////////////////////////////////////////////////
//                          CHANNELIZER 8x8                                  //
///////////////////////////////////////////////////////////////////////////////

// T is input type
// U is output type
template<typename T, typename U, int M = 8, int WARPSIZE = 32>
__global__ void _cupy_channelizer_8x8( const int n_chans,
                                       const int n_taps,
                                       const int n_pts,
                                       const T *__restrict__ x,
                                       const T *__restrict__ h,
                                       U *__restrict__ y ) {
    __shared__ T s_mem[M][M];
    
    const auto block { cg::this_thread_block( ) };
    const auto tile_32 { cg::tiled_partition<WARPSIZE>( block ) };
    const auto tile { cg::tiled_partition<M>( tile_32 ) };

    const auto btx { blockIdx.x * blockDim.x + threadIdx.x };

    const auto tx { threadIdx.x };
    const auto ty { threadIdx.y };

    // Initialize shared memory
    // Evaluate type at compile-time
    if ( btx < n_chans && ty < n_taps ) {
        if constexpr ( std::is_same_v<T, thrust::complex<float>> || std::is_same_v<T, thrust::complex<double>> ) {
            s_mem[tx][ty] = thrust::conj( h[ty * n_chans + btx] );
        } else {
            s_mem[tx][ty] = h[ty * n_chans + btx];
        }
    } else {
        if constexpr ( std::is_same_v<T, thrust::complex<float>> || std::is_same_v<T, thrust::complex<double>> ) {
            s_mem[tx][ty] = T( 0.0, 0.0 );
        } else {
            s_mem[tx][ty] = 0.0;
        }
    }

    block.sync( );

    T local_h { s_mem[ty][tx] };

    for ( auto bid = blockIdx.y; bid < n_pts; bid += gridDim.y ) {
        block.sync( );
        // Load data
        if ( bid >= n_taps ) {
            if ( btx < n_chans && ty < n_taps ) {
                if constexpr ( std::is_same_v<T, thrust::complex<float>> ||
                               std::is_same_v<T, thrust::complex<double>> ) {
                    s_mem[tx][( n_taps - 1 ) - ty] =
                        thrust::conj( x[( ( bid - n_taps + 1 ) + ty ) * n_chans + ( n_chans - 1 - btx )] );
                } else {
                    s_mem[tx][( n_taps - 1 ) - ty] = x[( ( bid - n_taps + 1 ) + ty ) * n_chans + ( n_chans - 1 - btx )];
                }
            }
        } else {
            if ( btx < n_chans && ty <= bid ) {
                if constexpr ( std::is_same_v<T, thrust::complex<float>> ||
                               std::is_same_v<T, thrust::complex<double>> ) {
                    s_mem[tx][bid - ty] = thrust::conj( x[ty * n_chans + ( n_chans - 1 - btx )] );
                } else {
                    s_mem[tx][bid - ty] = x[ty * n_chans + ( n_chans - 1 - btx )];
                }
            } else {
                if constexpr ( std::is_same_v<T, thrust::complex<float>> ||
                               std::is_same_v<T, thrust::complex<double>> ) {
                    s_mem[tx][ty] = T( 0.0, 0.0 );
                } else {
                    s_mem[tx][ty] = 0.0;
                }
            }
        }

        block.sync( );

        T local_reg { s_mem[ty][tx] };

        T temp {};
        U vv {};

        // Perform compute
        if ( ( blockIdx.x * M + ty ) < n_chans ) {
            temp = local_h * local_reg;
            if constexpr ( std::is_same_v<T, thrust::complex<float>> || std::is_same_v<T, thrust::complex<double>> ) {
                vv.real( cg::reduce( tile, temp.real( ), cg::plus<typename U::value_type>( ) ) );
                vv.imag( cg::reduce( tile, temp.imag( ), cg::plus<typename U::value_type>( ) ) );
            } else {
                vv.real( cg::reduce( tile, temp, cg::plus<typename U::value_type>( ) ) );
            }
        }

        // Store output
        if ( tx == 0 && ( blockIdx.x * M + ty ) < n_chans ) {
            y[bid * n_chans + ( blockIdx.x * M + ty )] = vv;
        }
    }
}
"""  # NOQA


@_util.memoize(for_each_device=True)
def _get_channelizer_8x8_module():
    return cupy.RawModule(
        code=_CHANNELIZER_8X8_KERNEL, options=('-std=c++17',),
        name_expressions=[
            '_cupy_channelizer_8x8<float,thrust::complex<float>>',
            '_cupy_channelizer_8x8<thrust::complex<float>,thrust::complex<float>>',
            '_cupy_channelizer_8x8<double,thrust::complex<double>>',
            '_cupy_channelizer_8x8<thrust::complex<double>,thrust::complex<double>>'],
    )


_CHANNELIZER_16X16_KERNEL = _CHANNELIZER_KERNEL_PREAMBLE + r"""
///////////////////////////////////////////////////////////////////////////////
//                        CHANNELIZER 16x16                                  //
///////////////////////////////////////////////////////////////////////////////

// T is input type
// U is output type
template<typename T, typename U, int M = 16, int WARPSIZE = 32>
__global__ void _cupy_channelizer_16x16( const int n_chans,
                                         const int n_taps,
                                         const int n_pts,
                                         const T *__restrict__ x,
                                         const T *__restrict__ h,
                                         U *__restrict__ y ) {
    __shared__ T s_mem[M][M];
    
    const auto block { cg::this_thread_block( ) };
    const auto tile_32 { cg::tiled_partition<WARPSIZE>( block ) };
    const auto tile { cg::tiled_partition<M>( tile_32 ) };

    const auto btx { blockIdx.x * blockDim.x + threadIdx.x };

    const auto tx { threadIdx.x };
    const auto ty { threadIdx.y };

    // Initialize shared memory
    // Evaluate type at compile-time
    if ( btx < n_chans && ty < n_taps ) {
        if constexpr ( std::is_same_v<T, thrust::complex<float>> || std::is_same_v<T, thrust::complex<double>> ) {
            s_mem[tx][ty] = thrust::conj( h[ty * n_chans + btx] );
        } else {
            s_mem[tx][ty] = h[ty * n_chans + btx];
        }
    } else {
        if constexpr ( std::is_same_v<T, thrust::complex<float>> || std::is_same_v<T, thrust::complex<double>> ) {
            s_mem[tx][ty] = T( 0.0, 0.0 );
        } else {
            s_mem[tx][ty] = 0.0;
        }
    }

    block.sync( );

    T local_h { s_mem[ty][tx] };

    for ( auto bid = blockIdx.y; bid < n_pts; bid += gridDim.y ) {
        block.sync( );
        // Load data
        if ( bid >= n_taps ) {
            if ( btx < n_chans && ty < n_taps ) {
                if constexpr ( std::is_same_v<T, thrust::complex<float>> ||
                               std::is_same_v<T, thrust::complex<double>> ) {
                    s_mem[tx][( n_taps - 1 ) - ty] =
                        thrust::conj( x[( ( bid - n_taps + 1 ) + ty ) * n_chans + ( n_chans - 1 - btx )] );
                } else {
                    s_mem[tx][( n_taps - 1 ) - ty] = x[( ( bid - n_taps + 1 ) + ty ) * n_chans + ( n_chans - 1 - btx )];
                }
            }
        } else {
            if ( btx < n_chans && ty <= bid ) {
                if constexpr ( std::is_same_v<T, thrust::complex<float>> ||
                               std::is_same_v<T, thrust::complex<double>> ) {
                    s_mem[tx][bid - ty] = thrust::conj( x[ty * n_chans + ( n_chans - 1 - btx )] );
                } else {
                    s_mem[tx][bid - ty] = x[ty * n_chans + ( n_chans - 1 - btx )];
                }
            } else {
                if constexpr ( std::is_same_v<T, thrust::complex<float>> ||
                               std::is_same_v<T, thrust::complex<double>> ) {
                    s_mem[tx][ty] = T( 0.0, 0.0 );
                } else {
                    s_mem[tx][ty] = 0.0;
                }
            }
        }

        block.sync( );

        T local_reg { s_mem[ty][tx] };

        T temp {};
        U vv {};

        // Perform compute
        if ( ( blockIdx.x * M + ty ) < n_chans ) {
            temp = local_h * local_reg;
            if constexpr ( std::is_same_v<T, thrust::complex<float>> || std::is_same_v<T, thrust::complex<double>> ) {
                vv.real( cg::reduce( tile, temp.real( ), cg::plus<typename U::value_type>( ) ) );
                vv.imag( cg::reduce( tile, temp.imag( ), cg::plus<typename U::value_type>( ) ) );
            } else {
                vv.real( cg::reduce( tile, temp, cg::plus<typename U::value_type>( ) ) );
            }
        }

        // Store output
        if ( tx == 0 && ( blockIdx.x * M + ty ) < n_chans ) {
            y[bid * n_chans + ( blockIdx.x * M + ty )] = vv;
        }
    }
}
"""  # NOQA


@_util.memoize(for_each_device=True)
def _get_channelizer_16x16_module():
    return cupy.RawModule(
        code=_CHANNELIZER_16X16_KERNEL, options=('-std=c++17',),
        name_expressions=[
            '_cupy_channelizer_16x16<float,thrust::complex<float>>',
            '_cupy_channelizer_16x16<thrust::complex<float>,thrust::complex<float>>',
            '_cupy_channelizer_16x16<double,thrust::complex<double>>',
            '_cupy_channelizer_16x16<thrust::complex<double>,thrust::complex<double>>'],
    )


_CHANNELIZER_32X32_KERNEL = _CHANNELIZER_KERNEL_PREAMBLE + r"""
///////////////////////////////////////////////////////////////////////////////
//                        CHANNELIZER 32x32                                  //
///////////////////////////////////////////////////////////////////////////////

// T is input type
// U is output type
template<typename T, typename U, int M = 32, int WARPSIZE = 32>
__global__ void _cupy_channelizer_32x32( const int n_chans,
                                         const int n_taps,
                                         const int n_pts,
                                         const T *__restrict__ x,
                                         const T *__restrict__ h,
                                         U *__restrict__ y ) {
    __shared__ T s_mem[M][M];
                                             
    const auto block { cg::this_thread_block( ) };
    const auto tile { cg::tiled_partition<WARPSIZE>( block ) };

    const auto btx { blockIdx.x * blockDim.x + threadIdx.x };

    const auto tx { threadIdx.x };
    const auto ty { threadIdx.y };

    // Initialize shared memory
    // Evaluate type at compile-time
    if ( btx < n_chans && ty < n_taps ) {
        if constexpr ( std::is_same_v<T, thrust::complex<float>> || std::is_same_v<T, thrust::complex<double>> ) {
            s_mem[tx][ty] = thrust::conj( h[ty * n_chans + btx] );
        } else {
            s_mem[tx][ty] = h[ty * n_chans + btx];
        }
    } else {
        if constexpr ( std::is_same_v<T, thrust::complex<float>> || std::is_same_v<T, thrust::complex<double>> ) {
            s_mem[tx][ty] = T( 0.0, 0.0 );
        } else {
            s_mem[tx][ty] = 0.0;
        }
    }

    block.sync( );

    T local_h { s_mem[ty][tx] };

    for ( auto bid = blockIdx.y; bid < n_pts; bid += gridDim.y ) {
        block.sync( );
        // Load data
        if ( bid >= n_taps ) {
            if ( btx < n_chans && ty < n_taps ) {
                if constexpr ( std::is_same_v<T, thrust::complex<float>> ||
                               std::is_same_v<T, thrust::complex<double>> ) {
                    s_mem[tx][( n_taps - 1 ) - ty] =
                        thrust::conj( x[( ( bid - n_taps + 1 ) + ty ) * n_chans + ( n_chans - 1 - btx )] );
                } else {
                    s_mem[tx][( n_taps - 1 ) - ty] = x[( ( bid - n_taps + 1 ) + ty ) * n_chans + ( n_chans - 1 - btx )];
                }
            }
        } else {
            if ( btx < n_chans && ty <= bid ) {
                if constexpr ( std::is_same_v<T, thrust::complex<float>> ||
                               std::is_same_v<T, thrust::complex<double>> ) {
                    s_mem[tx][bid - ty] = thrust::conj( x[ty * n_chans + ( n_chans - 1 - btx )] );
                } else {
                    s_mem[tx][bid - ty] = x[ty * n_chans + ( n_chans - 1 - btx )];
                }
            } else {
                if constexpr ( std::is_same_v<T, thrust::complex<float>> ||
                               std::is_same_v<T, thrust::complex<double>> ) {
                    s_mem[tx][ty] = T( 0.0, 0.0 );
                } else {
                    s_mem[tx][ty] = 0.0;
                }
            }
        }

        block.sync( );

        T local_reg { s_mem[ty][tx] };

        T temp {};
        U vv {};

        // Perform compute
        if ( ( blockIdx.x * M + ty ) < n_chans ) {
            temp = local_h * local_reg;
            if constexpr ( std::is_same_v<T, thrust::complex<float>> || std::is_same_v<T, thrust::complex<double>> ) {
                vv.real( cg::reduce( tile, temp.real( ), cg::plus<typename U::value_type>( ) ) );
                vv.imag( cg::reduce( tile, temp.imag( ), cg::plus<typename U::value_type>( ) ) );
            } else {
                vv.real( cg::reduce( tile, temp, cg::plus<typename U::value_type>( ) ) );
            }
        }

        // Store output
        if ( tx == 0 && ( blockIdx.x * M + ty ) < n_chans ) {
            y[bid * n_chans + ( blockIdx.x * M + ty )] = vv;
        }
    }
}
"""  # NOQA


@_util.memoize(for_each_device=True)
def _get_channelizer_32x32_module():
    return cupy.RawModule(
        code=_CHANNELIZER_32X32_KERNEL, options=('-std=c++17',),
        name_expressions=[
            '_cupy_channelizer_32x32<float,thrust::complex<float>>',
            '_cupy_channelizer_32x32<thrust::complex<float>,thrust::complex<float>>',
            '_cupy_channelizer_32x32<double,thrust::complex<double>>',
            '_cupy_channelizer_32x32<thrust::complex<double>,thrust::complex<double>>'],
    )


def _check_supported_type(np_type, k_type):
    if np_type not in _SUPPORTED_TYPES:
        raise ValueError(
            "Datatype {} not found for '{}'".format(np_type, k_type))


def _get_numSM():
    device_id = cupy.cuda.Device()
    return device_id.attributes["MultiProcessorCount"]


def _channelizer(x, h, y, n_chans, n_taps, n_pts):
    #  Blocks per grid sized for 2048 threads per SM
    np_type = str(x.dtype) + "_" + str(y.dtype)

    if n_taps <= 8:
        k_type = "channelizer_8x8"
        _check_supported_type(np_type, k_type)
        mod = _get_channelizer_8x8_module()
        threadsperblock = (8, 8)
        blockspergrid = ((n_chans + 7) // 8, _get_numSM() * 32)
    elif n_taps <= 16:
        k_type = "channelizer_16x16"
        _check_supported_type(np_type, k_type)
        mod = _get_channelizer_16x16_module()
        threadsperblock = (16, 16)
        blockspergrid = ((n_chans + 15) // 16, _get_numSM() * 8)
    elif n_taps <= 32:
        k_type = "channelizer_32x32"
        _check_supported_type(np_type, k_type)
        mod = _get_channelizer_32x32_module()
        threadsperblock = (32, 32)
        blockspergrid = ((n_chans + 31) // 32, _get_numSM() * 2)
    else:
        raise AssertionError()
    x_typename = get_typename(x.dtype)
    y_typename = get_typename(y.dtype)
    kernel_name = f'_cupy_{k_type}<{x_typename},{y_typename}>'
    kernel = mod.get_function(kernel_name)

    args = (n_chans, n_taps, n_pts, x, h, y)
    kernel(blockspergrid, threadsperblock, args)


def channelize_poly(x, h, n_chans):
    """
    Polyphase channelize signal into n channels

    Parameters
    ----------
    x : array_like
        The input data to be channelized
    h : array_like
        The 1-D input filter; will be split into n
        channels of int number of taps
    n_chans : int
        Number of channels for channelizer

    Returns
    -------
    yy : channelized output matrix

    Notes
    -----
    Currently only supports simple channelizer where channel
    spacing is equivalent to the number of channels used (zero overlap).
    Number of filter taps (len of filter / n_chans) must be <=32.

    """
    dtype = cupy.promote_types(x.dtype, h.dtype)

    x = cupy.asarray(x, dtype=dtype)
    h = cupy.asarray(h, dtype=dtype)

    # number of taps in each h_n filter
    n_taps = int(len(h) / n_chans)
    if n_taps > 32:
        raise NotImplementedError(
            "The number of calculated taps ({}) in  \
            each filter is currently capped at 32. Please reduce filter \
                length or number of channels".format(
                n_taps
            )
        )

    # number of outputs
    n_pts = int(len(x) / n_chans)

    if x.dtype == cupy.float32 or x.dtype == cupy.complex64:
        y = cupy.empty((n_pts, n_chans), dtype=cupy.complex64)
    elif x.dtype == cupy.float64 or x.dtype == cupy.complex128:
        y = cupy.empty((n_pts, n_chans), dtype=cupy.complex128)
    else:
        raise NotImplementedError(
            "Data type ({}) not allowed.".format(x.dtype))

    _channelizer(x, h, y, n_chans, n_taps, n_pts)

    return cupy.conj(cupy.fft.fft(y)).T



import numpy as np
import pytest

import cupy
from cupy import testing
import cupyx.scipy.signal

try:
    import scipy.signal  # NOQA
except ImportError:
    pass


@testing.with_requires("scipy")
class TestUniqueRoots:

    @testing.numpy_cupy_allclose(scipy_name='scp')
    def test_real_no_repeat(self, xp, scp):
        p = xp.asarray([-1.0, -0.5, 0.3, 1.2, 10.0])
        unique, multiplicity = scp.signal.unique_roots(p)
        return unique, multiplicity

    @pytest.mark.parametrize('rtype', ['min', 'max', 'avg'])
    @testing.numpy_cupy_allclose(scipy_name='scp')
    def test_real_repeat(self, xp, scp, rtype):
        p = xp.asarray([-1.0, -0.95, -0.89, -0.8, 0.5, 1.0, 1.05])

        unique, multiplicity = scp.signal.unique_roots(
            p, tol=1e-1, rtype=rtype)
        return unique, multiplicity

    @testing.numpy_cupy_allclose(scipy_name='scp')
    def test_complex_no_repeat(self, xp, scp):
        p = xp.asarray([-1.0, 1.0j, 0.5 + 0.5j, -1.0 - 1.0j, 3.0 + 2.0j])
        unique, multiplicity = scp.signal.unique_roots(p)
        return unique, multiplicity

    @pytest.mark.parametrize('rtype', ['min', 'max', 'avg'])
    @testing.numpy_cupy_allclose(scipy_name='scp')
    def test_complex_repeat(self, xp, scp, rtype):
        p = xp.asarray([-1.0, -1.0 + 0.05j, -0.95 + 0.15j, -0.90 + 0.15j, 0.0,
                        0.5 + 0.5j, 0.45 + 0.55j])

        unique, multiplicity = scp.signal.unique_roots(
            p, tol=1e-1, rtype='min')
        return unique, multiplicity

    @testing.numpy_cupy_allclose(scipy_name='scp')
    def test_gh_4915(self, xp, scp):
        p = np.roots(np.convolve(np.ones(5), np.ones(5)))
        # true_roots = [-(-1)**(1/5), (-1)**(4/5), -(-1)**(3/5), (-1)**(2/5)]

        unique, multiplicity = scp.signal.unique_roots(xp.asarray(p))
        unique = xp.sort(unique)
        return unique, multiplicity

    @pytest.mark.parametrize('p', [[1.0, 1.0j, 1.0], [1, 1 + 2e-9, 1e-9 + 1j]])
    @testing.numpy_cupy_allclose(scipy_name='scp')
    def test_complex_roots_extra(self, xp, scp, p):
        p = xp.asarray(p)
        unique, multiplicity = scp.signal.unique_roots(p)
        return unique, multiplicity

    @testing.numpy_cupy_allclose(scipy_name='scp')
    def test_single_unique_root(self, xp, scp):
        np.random.seed(1234)
        p = np.random.rand(100) + 1j * np.random.rand(100)
        p = xp.asarray(p)

        unique, multiplicity = scp.signal.unique_roots(p, 2)
        return unique, multiplicity


rtypes = ('avg', 'mean', 'min', 'minimum', 'max', 'maximum')


@testing.with_requires("scipy")
class TestPartialFractionExpansion:

    def test_compute_factors(self):

        from cupyx.scipy.signal._polyutils import _compute_factors

        factors, poly = _compute_factors([1, 2, 3], [3, 2, 1])
        assert len(factors) == 3
        testing.assert_allclose(factors[0], [1, -7, 16, -12], atol=1e-15)
        testing.assert_allclose(factors[1], [1, -6, 12, -10, 3], atol=1e-15)
        testing.assert_allclose(
            factors[2], [1, -7, 19, -25, 16, -4], atol=1e-15)
        testing.assert_allclose(
            poly, [1, -10, 40, -82, 91, -52, 12], atol=1e-15)

        factors, poly = _compute_factors([1, 2, 3], [3, 2, 1],
                                         include_powers=True)
        assert len(factors) == 6
        testing.assert_allclose(
            factors[0], [1,  -9,  31, -51,  40, -12], atol=1e-15)
        testing.assert_allclose(
            factors[1], [1,  -8,  23, -28,  12], atol=1e-15)
        testing.assert_allclose(factors[2], [1,  -7,  16, -12], atol=1e-15)
        testing.assert_allclose(
            factors[3], [1,  -8,  24, -34,  23,  -6], atol=1e-15)
        testing.assert_allclose(
            factors[4], [1,  -6,  12, -10,   3], atol=1e-15)
        testing.assert_allclose(
            factors[5], [1,  -7,  19, -25,  16,  -4], atol=1e-15)

        testing.assert_allclose(
            poly, [1, -10,  40, -82,  91, -52,  12], atol=1e-15)

    def test_group_poles(self):
        _group_poles = cupyx.scipy.signal.unique_roots

        unique, multiplicity = _group_poles(
            cupy.asarray([1.0, 1.001, 1.003, 2.0, 2.003, 3.0]), 0.1, 'min')
        testing.assert_allclose(unique, [1.0, 2.0, 3.0], atol=1e-15)
        testing.assert_allclose(multiplicity, [3, 2, 1], atol=1e-15)

    @pytest.mark.parametrize('ba',
                             [([5, 3, -2, 7], [-4, 0, 8, 3]),
                              ([-4, 8], [1, 6, 8]),
                                 ([4, 1], [1, -1, -2]),
                                 ([4, 3], [2, -3.4, 1.98, -0.406]),
                                 ([2, 1], [1, 5, 8, 4]),
                                 ([3, -1.1, 0.88, -2.396, 1.348],
                                  [1, -0.7, -0.14, 0.048]),
                                 ([1], [1, 2, -3]),
                                 ([1, 0, -5], [1, 0, 0, 0, -1]),
                                 ([3, 8, 6], [1, 3, 3, 1]),
                                 ([3, -1], [1, -3, 2]),
                                 ([2, 3, -1], [1, -3, 2]),
                                 ([7, 2, 3, -1], [1, -3, 2]),
                                 ([2, 3, -1], [1, -3, 4, -2]),
                              ])
    @testing.numpy_cupy_allclose(scipy_name='scp')
    def test_residue_general(self, xp, scp, ba):
        # Test are taken from issue #4464, note that poles in scipy are
        # in increasing by absolute value order, opposite to MATLAB.
        b, a = map(xp.asarray, ba)
        r, p, k = scp.signal.residue(b, a)
        return r, p, k

    @pytest.mark.parametrize('ba',
                             [([5, 3, -2, 7], [-4, 0, 8, 3]),
                              ([0, 5, 3, -2, 7], [-4, 0, 8, 3]),
                                 ([5, 3, -2, 7], [0, -4, 0, 8, 3]),
                                 ([0, 0, 5, 3, -2, 7], [0, 0, 0, -4, 0, 8, 3]),
                              ])
    @testing.numpy_cupy_allclose(scipy_name='scp')
    def test_residue_leading_zeros(self, xp, scp, ba):
        # Leading zeros in numerator or denominator must not affect the answer.
        ba = map(xp.asarray, ba)
        r0, p0, k0 = scp.signal.residue(*ba)
        return r0, p0, k0

    @pytest.mark.parametrize('ba',
                             [([0, 0], [1, 6, 8]),
                              (0, 1)
                              ])
    @testing.numpy_cupy_allclose(scipy_name='scp')
    def test_residue_degenerate(self, xp, scp, ba):
        # Several tests for zero numerator and denominator.
        ba = map(xp.asarray, ba)
        r, p, k = scp.signal.residue(*ba)
        return r, p, k

    @pytest.mark.parametrize('ba',
                             [([1, 6, 6, 2], [1, -(2 + 1j), (1 + 2j), -1j]),
                              ([1, 2, 1], [1, -1, 0.3561]),
                                 ([1, -1], [1, -5, 6]),
                                 ([2, 3, 4], [1, 3, 3, 1]),
                                 ([1, -10, -4, 4], [2, -2, -4]),
                                 ([18], [18, 3, -4, -1]),
                                 ([2, 3], np.polymul([1, -1/2], [1, 1/4])),
                                 ([1, -2, 1], [1, -1]),
                                 (1, [1, -1j]),
                                 (1, [1, -1, 0.25]),
                                 ([1, 6, 2], [1, -2, 1]),
                                 ([6, 2], [1, -2, 1]),
                                 ([1, 6, 6, 2], [1, -2, 1]),
                                 ([1, 0, 1], [1, 0, 0, 0, 0, -1])
                              ])
    @testing.numpy_cupy_allclose(scipy_name='scp')
    def test_residuez_general(self, xp, scp, ba):
        if (
                ba == ([1, 0, 1], [1, 0, 0, 0, 0, -1]) and
                np.__version__ >= np.lib.NumpyVersion('2.0.0')
        ):
            return xp.array([])  # Skip

        ba = map(xp.asarray, ba)
        r, p, k = scp.signal.residuez(*ba)
        return r, p, k

    @pytest.mark.parametrize('ba',
                             [([5, 3, -2, 7], [-4, 0, 8, 3]),
                              ([5, 3, -2, 7, 0], [-4, 0, 8, 3]),
                                 ([5, 3, -2, 7], [-4, 0, 8, 3, 0]),
                                 ([5, 3, -2, 7, 0, 0], [-4, 0, 8, 3, 0, 0, 0])
                              ])
    @testing.numpy_cupy_allclose(scipy_name='scp')
    def test_residuez_trailing_zeros(self, xp, scp, ba):
        # Trailing zeros in numerator or denominator must not affect the
        # answer.
        ba = map(xp.asarray, ba)
        r, p, k = scp.signal.residuez(*ba)
        return r, p, k

    @pytest.mark.parametrize('ba',
                             [([0, 0], [1, 6, 8]),
                              (0, 1)
                              ])
    @testing.numpy_cupy_allclose(scipy_name='scp')
    def test_residuez_degenerate(self, xp, scp, ba):
        ba = map(xp.asarray, ba)
        r, p, k = scp.signal.residuez(*ba)
        return r, p, k

    @pytest.mark.parametrize('rtype', rtypes)
    @testing.numpy_cupy_allclose(scipy_name='scp', atol=1e-15)
    def test_inverse_unique_roots_different_rtypes(self, xp, scp, rtype):
        # This test was inspired by github issue 2496.
        r = xp.asarray([3 / 10, -1 / 6, -2 / 15])
        p = xp.asarray([0, -2, -5])
        k = xp.asarray([])

        # With the default tolerance, the rtype does not matter
        # for this example.
        b, a = scp.signal.invres(r, p, k, rtype=rtype)
        return b, a

    @pytest.mark.parametrize('rtype', rtypes)
    @testing.numpy_cupy_allclose(scipy_name='scp', atol=1e-15)
    def test_inverse_unique_roots_different_rtypes_2(self, xp, scp, rtype):
        # This test was inspired by github issue 2496.
        r = xp.asarray([3 / 10, -1 / 6, -2 / 15])
        p = xp.asarray([0, -2, -5])
        k = xp.asarray([])

        b, a = scp.signal.invresz(r, p, k, rtype=rtype)
        return b, a

    @pytest.mark.parametrize('rtype', rtypes)
    @testing.numpy_cupy_allclose(scipy_name='scp', atol=1e-15)
    def test_inverse_repeated_roots_different_rtypes(self, xp, scp, rtype):
        r = xp.asarray([3 / 20, -7 / 36, -1 / 6, 2 / 45])
        p = xp.asarray([0, -2, -2, -5])
        k = xp.asarray([])
        b, a = scp.signal.invres(r, p, k, rtype=rtype)
        return b, a

    @pytest.mark.parametrize('rtype', rtypes)
    @testing.numpy_cupy_allclose(scipy_name='scp', atol=1e-15)
    def test_inverse_repeated_roots_different_rtypes_2(self, xp, scp, rtype):
        r = xp.asarray([3 / 20, -7 / 36, -1 / 6, 2 / 45])
        p = xp.asarray([0, -2, -2, -5])
        k = xp.asarray([])
        b, a = scp.signal.invresz(r, p, k, rtype=rtype)
        return b, a

    @testing.numpy_cupy_allclose(scipy_name='scp', atol=1e-15)
    def test_invresz_one_coefficient_bug(self, xp, scp):
        # Regression test for issue in gh-4646.
        r = xp.asarray([1])
        p = xp.asarray([2])
        k = xp.asarray([0])
        b, a = scp.signal.invresz(r, p, k)
        return b, a

    @pytest.mark.parametrize('rpk',
                             [([1], [1], []),
                              ([1 - 1j, 2, 0.5 - 3j], [1, 0.5j, 1 + 1j], []),
                                 ([0.5, 1], [1 - 1j, 2 + 2j], [1, 2, 3]),
                                 ([-1, 2, 1j, 3 - 1j, 4, -2],
                                  [-1, 2 - 1j, 2 - 1j, 3, 3, 3], []),
                                 ([-1, 1j], [1, 1], [1, 2])
                              ])
    @testing.numpy_cupy_allclose(scipy_name='scp', atol=1e-15)
    def test_invres(self, xp, scp, rpk):
        r, p, k = map(xp.asarray, rpk)
        b, a = scp.signal.invres(r, p, k)
        return b, a

    @pytest.mark.parametrize('rpk',
                             [([1], [1], []),
                              ([1 - 1j, 2, 0.5 - 3j], [1, 0.5j, 1 + 1j], []),
                                 ([0.5, 1], [1 - 1j, 2 + 2j], [1, 2, 3]),
                                 ([-1, 2, 1j, 3 - 1j, 4, -2],
                                  [-1, 2 - 1j, 2 - 1j, 3, 3, 3], []),
                                 ([-1, 1j], [1, 1], [1, 2])
                              ])
    @testing.numpy_cupy_allclose(scipy_name='scp', atol=1e-15)
    def test_invresz(self, xp, scp, rpk):
        r, p, k = map(xp.asarray, rpk)
        b, a = scp.signal.invresz(r, p, k)
        return b, a

    @testing.numpy_cupy_allclose(scipy_name='scp', atol=1e-15)
    def test_inverse_scalar_arguments(self, xp, scp):
        b, a = scp.signal.invres(1, 1, 1)
        return b, a

    @testing.numpy_cupy_allclose(scipy_name='scp', atol=1e-15)
    def test_inverse_scalar_arguments_2(self, xp, scp):
        b, a = scp.signal.invresz(1, 1, 1)
        return b, a



import itertools
import unittest

from cupy import testing
from cupy_tests.core_tests.fusion_tests import fusion_utils


def _permutate_shapes(shapes_list):
    # Permutates input shapes
    permutated_shapes_set = set()
    for shapes in shapes_list:
        for permutated_shapes in itertools.permutations(shapes):
            permutated_shapes_set.add(permutated_shapes)
    return list(permutated_shapes_set)


@testing.parameterize(*testing.product({
    'shapes': _permutate_shapes([
        # Same shapes
        ((1,), (1,)),
        ((3, 4), (3, 4)),

        # Broadcast
        ((10,), (1,)),
        ((3, 4), (3, 1)),
        ((3, 4), (1, 4)),
        ((3, 4), (4,)),
        ((3, 4), (1, 1)),
        ((3, 4), (1,)),
        ((2, 3, 4), (1, 1, 1)),
        ((3, 1), (1, 4)),
        ((2, 1, 4), (3, 1)),

        # TODO(asi1024): Fix testing.shaped_random to support 0-dim array.
        # # 0-dim shape
        # ((), ()),
        # ((1,), ()),
        # ((3,), ()),
        # ((2, 3), ()),

        # 0-size shape
        ((0,), (0,)),
        ((0,), (1,)),
        ((2, 0, 3), (2, 0, 3)),
        ((2, 0, 3), (0, 1)),

        # Large case
        ((256, 256), (256,)),
        ((256, 256), (256, 1)),
    ])
}))
class TestFusionBroadcast(unittest.TestCase):

    def generate_inputs(self, xp):
        shape1, shape2 = self.shapes
        x = testing.shaped_random(shape1, xp, 'int64', scale=10, seed=0)
        y = testing.shaped_random(shape2, xp, 'int64', scale=10, seed=1)
        return (x, y), {}

    @fusion_utils.check_fusion()
    def test_broadcast(self, xp):
        return lambda x, y: x + y

    # TODO(asi1024): Uncomment after replace fusion implementation.

    # @fusion_utils.check_fusion(accept_error=ValueError)
    # def test_broadcast_inplace(self, xp):
    #     def impl(x, y):
    #         x += y
    #     return impl


@testing.parameterize(*testing.product({
    'shapes': _permutate_shapes([
        ((2,), (3,)),
        ((2,), (0,)),
        ((3, 2), (3, 3)),
        ((3, 2), (2, 2)),
        ((3,), (1, 2)),
    ])
}))
class TestFusionBroadcastInvalid(unittest.TestCase):

    def generate_inputs(self, xp):
        shape1, shape2 = self.shapes
        x = testing.shaped_random(shape1, xp, 'int64', scale=10, seed=0)
        y = testing.shaped_random(shape2, xp, 'int64', scale=10, seed=1)
        return (x, y), {}

    @fusion_utils.check_fusion(accept_error=ValueError)
    def test_broadcast(self, xp):
        return lambda x, y: x + y

    @fusion_utils.check_fusion(accept_error=ValueError)
    def test_broadcast_inplace(self, xp):
        def impl(x, y):
            x += y
        return impl


class TestFusionParseInput(unittest.TestCase):

    def generate_inputs(self, xp):
        x = testing.shaped_random((3, 4), xp, 'int64', scale=10, seed=0)
        return (x,), {}

    @fusion_utils.check_fusion()
    def test_add(self, xp):
        return lambda x: x + x

    # TODO(asi1024): Should fix cupy.ufunc
    @fusion_utils.check_fusion(accept_error=(ValueError, TypeError))
    def test_add_too_less_param(self, xp):
        return lambda x: xp.add(x)

    # TODO(asi1024): Should fix cupy.ufunc
    @fusion_utils.check_fusion(accept_error=(ValueError, TypeError))
    def test_add_too_much_param(self, xp):
        return lambda x: xp.add(x, x, x, x)

    @fusion_utils.check_fusion(accept_error=TypeError)
    def test_add_none(self, xp):
        return lambda x: x + None

    @fusion_utils.check_fusion(accept_error=TypeError)
    def test_add_object(self, xp):
        return lambda x: x + object()

    # TODO(asi1024): Should fix cupy.ufunc
    # @fusion_utils.check_fusion()
    # def test_add_out_none(self, xp):
    #     def impl(x):
    #         xp.add(x, x, None)
    #         return x
    #     return impl

    @fusion_utils.check_fusion()
    def test_add_kwargs_out_none(self, xp):
        def impl(x):
            xp.add(x, x, out=None)
        return impl

    # TODO(asi1024): Should fix cupy.ufunc
    # @fusion_utils.check_fusion(accept_error=ValueError)
    # def test_add_both_out_none(self, xp):
    #     def impl(x):
    #         xp.add(x, x, None, out=None)
    #     return impl

    @fusion_utils.check_fusion(accept_error=TypeError)
    def test_add_out_object(self, xp):
        def impl(x):
            xp.add(x, x, object())
            return x
        return impl

    @fusion_utils.check_fusion(accept_error=TypeError)
    def test_add_kwargs_out_object(self, xp):
        def impl(x):
            xp.add(x, x, out=object())
            return x
        return impl

    @fusion_utils.check_fusion()
    def test_divmod(self, xp):
        return lambda x: xp.divmod(x, x)


class TestFusionOutDtype(unittest.TestCase):

    def generate_inputs(self, xp, dtype1, dtype2):
        x = testing.shaped_random((3, 4), xp, dtype1, scale=10, seed=0)
        y = testing.shaped_random((3, 4), xp, dtype1, scale=10, seed=1)
        z = testing.shaped_random((3, 4), xp, dtype2, scale=10, seed=2)
        return (x, y, z), {}

    @testing.for_all_dtypes_combination(
        names=('dtype1', 'dtype2'), full=True, no_complex=True)
    @fusion_utils.check_fusion(accept_error=TypeError)
    @testing.with_requires('numpy>=1.13')
    def test_outarg(self, xp, dtype1, dtype2):
        def impl(x, y, z):
            xp.add(x, y, out=z)
            return z
        return impl


class TestFusionScalar(unittest.TestCase):

    def generate_inputs(self, xp, dtype1, dtype2):
        array = testing.shaped_random((3, 4), xp, dtype1, scale=10, seed=0)
        return (array,), {}

    @testing.for_all_dtypes_combination(names=('dtype1', 'dtype2'))
    @fusion_utils.check_fusion()
    def test_python_scalar_r(self, xp, dtype1, dtype2):
        def func(array):
            py_scalar = dtype2(1).item()
            return array + py_scalar

        return func

    @testing.for_all_dtypes_combination(names=('dtype1', 'dtype2'))
    @fusion_utils.check_fusion()
    def test_numpy_scalar_r(self, xp, dtype1, dtype2):
        def func(array):
            np_scalar = dtype2(1)
            return array + np_scalar

        return func

    @testing.for_all_dtypes_combination(names=('dtype1', 'dtype2'))
    @fusion_utils.check_fusion()
    def test_python_scalar_l(self, xp, dtype1, dtype2):
        def func(array):
            py_scalar = dtype2(1).item()
            return py_scalar + array

        return func

    @testing.for_all_dtypes_combination(names=('dtype1', 'dtype2'))
    @fusion_utils.check_fusion()
    def test_numpy_scalar_l(self, xp, dtype1, dtype2):
        def func(array):
            np_scalar = dtype2(1)
            return np_scalar + array

        return func

    def python_scalar_param_r(self, xp, dtype1, dtype2):
        array = testing.shaped_random((3, 4), xp, dtype1, scale=10, seed=0)
        py_scalar = dtype2(1).item()
        return (array, py_scalar), {}

    @testing.for_all_dtypes_combination(names=('dtype1', 'dtype2'))
    @fusion_utils.check_fusion(generate_inputs_name='python_scalar_param_r')
    def test_python_scalar_param_r(self, xp, dtype1, dtype2):
        def func(array, py_scalar):
            return array + py_scalar

        return func

    def python_scalar_param_l(self, xp, dtype1, dtype2):
        array = testing.shaped_random((3, 4), xp, dtype1, scale=10, seed=0)
        py_scalar = dtype2(1).item()
        return (py_scalar, array), {}

    @testing.for_all_dtypes_combination(names=('dtype1', 'dtype2'))
    @fusion_utils.check_fusion(generate_inputs_name='python_scalar_param_l')
    def test_python_scalar_param_l(self, xp, dtype1, dtype2):
        def func(py_scalar, array):
            return py_scalar + array

        return func

    def numpy_scalar_param_r(self, xp, dtype1, dtype2):
        array = testing.shaped_random((3, 4), xp, dtype1, scale=10, seed=0)
        np_scalar = dtype2(1)
        return (array, np_scalar), {}

    @testing.for_all_dtypes_combination(names=('dtype1', 'dtype2'))
    @fusion_utils.check_fusion(generate_inputs_name='numpy_scalar_param_r')
    def test_numpy_scalar_param_r(self, xp, dtype1, dtype2):
        def func(array, np_scalar):
            return array + np_scalar

        return func

    def numpy_scalar_param_l(self, xp, dtype1, dtype2):
        array = testing.shaped_random((3, 4), xp, dtype1, scale=10, seed=0)
        np_scalar = dtype2(1)
        return (np_scalar, array), {}

    @testing.for_all_dtypes_combination(names=('dtype1', 'dtype2'))
    @fusion_utils.check_fusion(generate_inputs_name='numpy_scalar_param_l')
    def test_numpy_scalar_param_l(self, xp, dtype1, dtype2):
        def func(np_scalar, array):
            return np_scalar + array

        return func

    def numpy_scalar_params_binop(self, xp, dtype1, dtype2):
        scalar1 = dtype1(1)
        scalar2 = dtype2(1)
        array = testing.shaped_random((3, 4), xp, 'int64', scale=10, seed=0)
        return (scalar1, scalar2, array), {}

    @testing.for_all_dtypes_combination(names=('dtype1', 'dtype2'))
    @fusion_utils.check_fusion(
        generate_inputs_name='numpy_scalar_params_binop')
    def test_numpy_scalar_params_binop(self, xp, dtype1, dtype2):
        def func(scalar1, scalar2, array):
            dtype = (scalar1 + scalar2).dtype
            return array.astype(dtype)

        return func

    @testing.for_all_dtypes_combination(names=('dtype1', 'dtype2'))
    @fusion_utils.check_fusion(
        generate_inputs_name='numpy_scalar_params_binop')
    def test_scalar_inplace_update(self, xp, dtype1, dtype2):
        def func(scalar1, scalar2, array):
            scalar1_copy = scalar1
            scalar1 += scalar2
            return array + scalar1 + scalar1_copy

        return func

    @testing.for_all_dtypes_combination(names=('dtype1', 'dtype2'))
    @fusion_utils.check_fusion(generate_inputs_name='numpy_scalar_param_r')
    def test_scalar_inplace_update_with_array(self, xp, dtype1, dtype2):
        def func(array, scalar):
            scalar += array
            return scalar

        return func

    @testing.for_all_dtypes_combination(names=('dtype1', 'dtype2'))
    @fusion_utils.check_fusion(generate_inputs_name='numpy_scalar_param_r')
    def test_scalar_and_constant_mix(self, xp, dtype1, dtype2):
        # Example from gh-8536
        def func(array, scalar):
            return 1.0 - (array * array) / (scalar * scalar)

        return func



import warnings

import cupy
from cupy._core import _routines_logic as _logic
from cupy._core import _fusion_thread_local
from cupy._sorting import search as _search
from cupy import _util


_setxorkernel = cupy._core.ElementwiseKernel(
    'raw T X, int64 len',
    'bool z',
    'z = (i == 0 || X[i] != X[i-1]) && (i == len - 1 || X[i] != X[i+1])',
    'setxorkernel'
)


def all(a, axis=None, out=None, keepdims=False):
    """Tests whether all array elements along a given axis evaluate to True.

    Parameters
    ----------
    a : cupy.ndarray
        Input array.
    axis : int or tuple of ints
        Along which axis to compute all.
        The flattened array is used by default.
    out : cupy.ndarray
        Output array.
    keepdims : bool
        If ``True``, the axis is remained as an axis of size one.

    Returns
    -------
    y : cupy.ndarray
        An array reduced of the input array along the axis.

    See Also
    --------
    numpy.all

    """
    if _fusion_thread_local.is_fusing():
        if keepdims:
            raise NotImplementedError(
                'cupy.all does not support `keepdims` in fusion yet.')
        return _fusion_thread_local.call_reduction(
            _logic.all, a, axis=axis, out=out)

    _util.check_array(a, arg_name='a')

    return a.all(axis=axis, out=out, keepdims=keepdims)


def any(a, axis=None, out=None, keepdims=False):
    """Tests whether any array elements along a given axis evaluate to True.

    Parameters
    ----------
    a : cupy.ndarray
        Input array.
    axis : int or tuple of ints
        Along which axis to compute all.
        The flattened array is used by default.
    out : cupy.ndarray
        Output array.
    keepdims : bool
        If ``True``, the axis is remained as an axis of size one.

    Returns
    -------
    y : cupy.ndarray
        An array reduced of the input array along the axis.

    See Also
    --------
    numpy.any

    """
    if _fusion_thread_local.is_fusing():
        if keepdims:
            raise NotImplementedError(
                'cupy.any does not support `keepdims` in fusion yet.')
        return _fusion_thread_local.call_reduction(
            _logic.any, a, axis=axis, out=out)

    _util.check_array(a, arg_name='a')

    return a.any(axis=axis, out=out, keepdims=keepdims)


def in1d(ar1, ar2, assume_unique=False, invert=False):
    """Tests whether each element of a 1-D array is also present in a second
    array.

    Returns a boolean array the same length as ``ar1`` that is ``True``
    where an element of ``ar1`` is in ``ar2`` and ``False`` otherwise.

    Parameters
    ----------
    ar1 : cupy.ndarray
        Input array.
    ar2 : cupy.ndarray
        The values against which to test each value of ``ar1``.
    assume_unique : bool, optional
        Ignored
    invert : bool, optional
        If ``True``, the values in the returned array
        are inverted (that is, ``False`` where an element of ``ar1`` is in
        ``ar2`` and ``True`` otherwise). Default is ``False``.

    Returns
    -------
    y : cupy.ndarray, bool
        The values ``ar1[in1d]`` are in ``ar2``.

    """
    # Ravel both arrays, behavior for the first array could be different
    ar1 = ar1.ravel()
    ar2 = ar2.ravel()
    if ar1.size == 0 or ar2.size == 0:
        if invert:
            return cupy.ones(ar1.shape, dtype=cupy.bool_)
        else:
            return cupy.zeros(ar1.shape, dtype=cupy.bool_)
    # Use brilliant searchsorted trick
    # https://github.com/cupy/cupy/pull/4018#discussion_r495790724
    ar2 = cupy.sort(ar2)
    return _search._exists_kernel(ar1, ar2, ar2.size, invert)


def intersect1d(arr1, arr2, assume_unique=False, return_indices=False):
    """Find the intersection of two arrays.
    Returns the sorted, unique values that are in both of the input arrays.

    Parameters
    ----------
    arr1, arr2 : cupy.ndarray
        Input arrays. Arrays will be flattened if they are not in 1D.
    assume_unique : bool
        By default, False. If set True, the input arrays will be
        assumed to be unique, which speeds up the calculation. If set True,
        but the arrays are not unique, incorrect results and out-of-bounds
        indices could result.
    return_indices : bool
       By default, False. If True, the indices which correspond to the
       intersection of the two arrays are returned.

    Returns
    -------
    intersect1d : cupy.ndarray
        Sorted 1D array of common and unique elements.
    comm1 : cupy.ndarray
        The indices of the first occurrences of the common values
        in `arr1`. Only provided if `return_indices` is True.
    comm2 : cupy.ndarray
        The indices of the first occurrences of the common values
        in `arr2`. Only provided if `return_indices` is True.

    See Also
    --------
    numpy.intersect1d

    """
    if not assume_unique:
        if return_indices:
            arr1, ind1 = cupy.unique(arr1, return_index=True)
            arr2, ind2 = cupy.unique(arr2, return_index=True)
        else:
            arr1 = cupy.unique(arr1)
            arr2 = cupy.unique(arr2)
    else:
        arr1 = arr1.ravel()
        arr2 = arr2.ravel()

    if not return_indices:
        mask = _search._exists_kernel(arr1, arr2, arr2.size, False)
        return arr1[mask]

    mask, v1 = _search._exists_and_searchsorted_kernel(
        arr1, arr2, arr2.size, False)
    int1d = arr1[mask]
    arr1_indices = cupy.flatnonzero(mask)
    arr2_indices = v1[mask]

    if not assume_unique:
        arr1_indices = ind1[arr1_indices]
        arr2_indices = ind2[arr2_indices]

    return int1d, arr1_indices, arr2_indices


def isin(element, test_elements, assume_unique=False, invert=False):
    """Calculates element in ``test_elements``, broadcasting over ``element``
    only. Returns a boolean array of the same shape as ``element`` that is
    ``True`` where an element of ``element`` is in ``test_elements`` and
    ``False`` otherwise.

    Parameters
    ----------
    element : cupy.ndarray
        Input array.
    test_elements : cupy.ndarray
        The values against which to test each
        value of ``element``. This argument is flattened if it is an
        array or array_like.
    assume_unique : bool, optional
        Ignored
    invert : bool, optional
        If ``True``, the values in the returned array
        are inverted, as if calculating element not in ``test_elements``.
        Default is ``False``.

    Returns
    -------
    y : cupy.ndarray, bool
        Has the same shape as ``element``. The values ``element[isin]``
        are in ``test_elements``.

    """
    return in1d(element, test_elements, assume_unique=assume_unique,
                invert=invert).reshape(element.shape)


def setdiff1d(ar1, ar2, assume_unique=False):
    """Find the set difference of two arrays. It returns unique
    values in `ar1` that are not in `ar2`.

    Parameters
    ----------
    ar1 : cupy.ndarray
        Input array
    ar2 : cupy.ndarray
        Input array for comparison
    assume_unique : bool
        By default, False, i.e. input arrays are not unique.
        If True, input arrays are assumed to be unique. This can
        speed up the calculation.

    Returns
    -------
    setdiff1d : cupy.ndarray
        Returns a 1D array of values in `ar1` that are not in `ar2`.
        It always returns a sorted output for unsorted input only
        if `assume_unique=False`.

    See Also
    --------
    numpy.setdiff1d

    """
    if assume_unique:
        ar1 = cupy.ravel(ar1)
    else:
        ar1 = cupy.unique(ar1)
        ar2 = cupy.unique(ar2)
    return ar1[in1d(ar1, ar2, assume_unique=True, invert=True)]


def setxor1d(ar1, ar2, assume_unique=False):
    """Find the set exclusive-or of two arrays.

    Parameters
    ----------
    ar1, ar2 : cupy.ndarray
        Input arrays. They are flattened if they are not already 1-D.
    assume_unique : bool
        By default, False, i.e. input arrays are not unique.
        If True, input arrays are assumed to be unique. This can
        speed up the calculation.

    Returns
    -------
    setxor1d : cupy.ndarray
        Return the sorted, unique values that are in only one
        (not both) of the input arrays.

    See Also
    --------
    numpy.setxor1d

    """
    if not assume_unique:
        ar1 = cupy.unique(ar1)
        ar2 = cupy.unique(ar2)

    aux = cupy.concatenate((ar1, ar2), axis=None)
    if aux.size == 0:
        return aux

    aux.sort()

    return aux[_setxorkernel(aux, aux.size,
                             cupy.zeros(aux.size, dtype=cupy.bool_))]


def union1d(arr1, arr2):
    """Find the union of two arrays.

    Returns the unique, sorted array of values that are in either of
    the two input arrays.

    Parameters
    ----------
    arr1, arr2 : cupy.ndarray
        Input arrays. They are flattened if they are not already 1-D.

    Returns
    -------
    union1d : cupy.ndarray
        Sorted union of the input arrays.

    See Also
    --------
    numpy.union1d

    """
    return cupy.unique(cupy.concatenate((arr1, arr2), axis=None))


def alltrue(a, axis=None, out=None, keepdims=False):
    warnings.warn('Please use `all` instead.', DeprecationWarning)
    return all(a, axis, out, keepdims)


def sometrue(a, axis=None, out=None, keepdims=False):
    warnings.warn('Please use `any` instead.', DeprecationWarning)
    return any(a, axis, out, keepdims)



import math

import cupy
import numpy
import pytest

import cupyx.scipy.special
from cupy import testing
from cupy.testing import (
    assert_array_equal,
    assert_array_almost_equal,
)
from cupy.testing import numpy_cupy_allclose

rtol = {'default': 1e-5, cupy.float64: 1e-12, cupy.complex128: 1e-12}


@testing.with_requires("scipy")
class TestLegendreFunctions:

    def test_lpmv_basic(self):
        # specific values tested in the SciPy test suite
        scp = cupyx.scipy
        lp = scp.special.lpmv(0, 2, 0.5)
        assert_array_almost_equal(lp, -0.125, 7)
        lp = scp.special.lpmv(0, 40, 0.001)
        assert_array_almost_equal(lp, 0.1252678976534484, 7)

        # XXX: this is outside the domain of the current implementation,
        #      so ensure it returns a NaN rather than a wrong answer.
        olderr = numpy.seterr(all="ignore")
        try:
            lp = scp.special.lpmv(-1, -1, 0.001)
        finally:
            numpy.seterr(**olderr)
        assert lp != 0 or cupy.isnan(lp)

    @pytest.mark.parametrize("order", [0, 1, 2, 3, 4])
    @pytest.mark.parametrize("degree", [0, 1, 2, 3, 4, 5, 10, 20, 30, 40, 50])
    @testing.for_dtypes("efd")
    @numpy_cupy_allclose(scipy_name="scp", atol=1e-12)
    def test_lpmv(self, xp, scp, dtype, order, degree):
        vals = xp.linspace(-1, 1, 100, dtype=dtype)
        return scp.special.lpmv(order, degree, vals)


@testing.with_requires("scipy<1.16")
class TestBasic:

    @testing.with_requires('scipy<1.15')
    @testing.for_dtypes("efd")
    @numpy_cupy_allclose(scipy_name="scp")
    def test_gammasgn(self, xp, scp, dtype):
        vals = xp.linspace(-4, 4, 100, dtype=dtype)
        return scp.special.gammasgn(vals)

    @testing.for_dtypes("efdFD")
    @numpy_cupy_allclose(scipy_name="scp", rtol=1e-5)
    def test_log1p_linspace(self, xp, scp, dtype):
        vals = xp.linspace(-100, 100, 1000, dtype=dtype)
        dtype = xp.dtype(dtype)
        if dtype.kind == 'c':
            # broadcast to mix large and small real vs. imaginary
            vals = vals[::10, xp.newaxis] + 1j * vals[xp.newaxis, ::10]
        return xp.abs(scp.special.log1p(vals))

    @testing.for_dtypes("efFdD")
    @numpy_cupy_allclose(scipy_name="scp", rtol=1e-5)
    def test_log1p_logspace(self, xp, scp, dtype):
        dtype = xp.dtype(dtype)
        vals = xp.logspace(-10, 10, 1000, dtype=dtype)
        if dtype.kind == 'c':
            # broadcast to mix large and small real vs. imaginary
            vals = vals[::10, xp.newaxis] + 1j * vals[xp.newaxis, ::10]
        return xp.abs(scp.special.log1p(vals))

    @testing.for_dtypes("efd")
    @numpy_cupy_allclose(scipy_name="scp", rtol=rtol)
    def test_log1p_path2(self, xp, scp, dtype):
        # test values for code path corresponding to range [1/sqrt(2), sqrt(2)]
        vals = xp.linspace(1 / math.sqrt(2), math.sqrt(2), 1000, dtype=dtype)
        return scp.special.log1p(vals)

    def test_log1p_real(self):
        log1p = cupyx.scipy.special.log1p
        inf = cupy.inf
        nan = cupy.nan
        assert_array_equal(log1p(0), 0.0)
        assert_array_equal(log1p(-1), -inf)
        assert_array_equal(log1p(-2), nan)
        assert_array_equal(log1p(inf), inf)

    @pytest.mark.parametrize("function", ["xlogy", "xlog1py"])
    @testing.for_dtypes("efdFD")
    @numpy_cupy_allclose(scipy_name="scp", rtol={'default': 1e-3,
                                                 cupy.float64: 1e-12})
    def test_xlogy(self, xp, scp, dtype, function):
        # only test with values > 0 to avoid NaNs
        x = xp.linspace(-100, 100, 1000, dtype=dtype)
        y = xp.linspace(0.001, 100, 1000, dtype=dtype)
        if x.dtype.kind == 'c':
            x -= 1j * x
            y += 1j * y
        return getattr(scp.special, function)(x, y)

    @pytest.mark.parametrize("function", ["xlogy", "xlog1py"])
    @testing.for_dtypes('efdFD')
    @numpy_cupy_allclose(scipy_name="scp", rtol={'default': 1e-3,
                                                 cupy.float64: 1e-12})
    def test_xlogy_zeros(self, xp, scp, dtype, function):
        # only test with values > 0 to avoid NaNs
        x = xp.zeros((1, 100), dtype=dtype)
        y = xp.linspace(-10, 10, 100, dtype=dtype)
        if y.dtype.kind == 'c':
            y += 1j * y
        return getattr(scp.special, function)(x, y)

    @pytest.mark.parametrize("function", ["xlogy", "xlog1py"])
    @testing.for_all_dtypes()
    def test_xlogy_nonfinite(self, dtype, function):
        func = getattr(cupyx.scipy.special, function)
        y = cupy.ones((5,), dtype=dtype)
        assert cupy.all(cupy.isnan(func(cupy.nan, y)))
        assert cupy.all(cupy.isnan(func(y, cupy.nan)))

    @testing.for_dtypes("efd")
    @numpy_cupy_allclose(scipy_name="scp", rtol=1e-6)
    def test_exp2(self, xp, scp, dtype):
        vals = xp.linspace(-100, 100, 200, dtype=dtype)
        return scp.special.exp2(vals)

    @testing.for_dtypes("efd")
    @numpy_cupy_allclose(scipy_name="scp", rtol=1e-6)
    def test_exp10(self, xp, scp, dtype):
        if xp.dtype(dtype).char == 'd':
            vals = xp.linspace(-100, 100, 100, dtype=dtype)
        else:
            # Note: comparisons start to fail outside this range
            #       np.finfo(np.float32).max is 3.4028235e+38
            vals = xp.linspace(-37, 37, 100, dtype=dtype)
        return scp.special.exp10(vals)

    @testing.for_dtypes("efdFD")
    @numpy_cupy_allclose(scipy_name="scp", rtol=1e-6)
    def test_expm1(self, xp, scp, dtype):
        vals = xp.linspace(-50, 50, 200, dtype=dtype)
        if xp.dtype(dtype).kind == 'c':
            # broadcast to mix small and large real and imaginary parts
            vals = vals[:, xp.newaxis] + 1j * vals[xp.newaxis, :]
        return scp.special.expm1(vals)

    @testing.for_dtypes("efd")
    @numpy_cupy_allclose(scipy_name="scp", rtol=1e-6)
    def test_cosm1(self, xp, scp, dtype):
        vals = xp.linspace(-50, 50, 200, dtype=dtype)
        return scp.special.cosm1(vals)

    @testing.for_dtypes("efd")
    @numpy_cupy_allclose(scipy_name="scp", rtol=1e-6)
    def test_cosm1_close_to_zero(self, xp, scp, dtype):
        vals = xp.linspace(-1e-8, 1e-8, 200, dtype=dtype)
        return scp.special.cosm1(vals)

    @testing.for_dtypes("efd")
    @numpy_cupy_allclose(scipy_name="scp", rtol=1e-6)
    def test_radian(self, xp, scp, dtype):
        tmp = xp.linspace(-100, 100, 10, dtype=dtype)
        d = tmp[:, xp.newaxis, xp.newaxis]
        m = tmp[xp.newaxis, : xp.newaxis]
        s = tmp[xp.newaxis, xp.newaxis, :]
        return scp.special.radian(d, m, s)

    @pytest.mark.parametrize('function', ['cosdg', 'sindg', 'tandg', 'cotdg'])
    @testing.for_dtypes("efd")
    @numpy_cupy_allclose(scipy_name="scp",
                         atol={'default': 1e-6, cupy.float64: 1e-12},
                         rtol=1e-6)
    def test_trig_degrees(self, xp, scp, dtype, function):
        vals = xp.linspace(-100, 100, 200, dtype=dtype)
        # test at exact multiples of 45 degrees
        vals = xp.concatenate((vals, xp.arange(-360, 361, 45, dtype=dtype)))
        return getattr(scp.special, function)(vals)

    @testing.for_dtypes("efd")
    @numpy_cupy_allclose(scipy_name="scp", rtol=1e-6)
    def test_cbrt(self, xp, scp, dtype):
        vals = xp.linspace(-100, 100, 200, dtype=dtype)
        return scp.special.cbrt(vals)

    # TODO: omit "e" since SciPy will promote to "f", but CuPy does not
    @testing.for_dtypes("fd")
    @numpy_cupy_allclose(scipy_name="scp", rtol=1e-6)
    def test_round(self, xp, scp, dtype):
        vals = xp.concatenate(
            (
                xp.linspace(-2, 2, 100, dtype=dtype),
                # test at half-integer locations
                xp.asarray([-2.5, -1.5, -0.5, 0.5, 1.5, 2.5], dtype=dtype),
            )
        )
        return scp.special.round(vals)

    @testing.with_requires("numpy<2.0")
    # Exclude 'e' here because of deficiency in the NumPy/SciPy
    # implementation for float16 dtype. This was also noted in
    # cupy_tests/math_tests/test_special.py
    @testing.for_dtypes("fd")
    @numpy_cupy_allclose(scipy_name="scp", rtol=1e-6, atol=1e-6)
    def test_sinc(self, xp, scp, dtype):
        vals = xp.linspace(-100, 100, 200, dtype=dtype)
        return scp.special.sinc(vals)

    @testing.with_requires("numpy<2.0")
    # TODO: Should we make all int dtypes convert to float64 and test for that?
    #       currently int8->float16, int16->float32, etc... but for
    #       numpy.sinc/scipy.special.sinc any int type becomes float64.
    @testing.for_dtypes("fd")
    @numpy_cupy_allclose(scipy_name="scp", rtol=1e-6, atol=1e-6)
    def test_sinc_integer_valued(self, xp, scp, dtype):
        vals = xp.arange(0, 10, dtype=dtype)
        return scp.special.sinc(vals)



from math import sqrt, pi

import cupy
import cupyx.scipy.signal as signal
from cupyx.scipy.signal._iir_filter_conversions import _cplxreal

from cupy import testing
from cupy.testing import assert_array_almost_equal

import numpy as np

import pytest
from pytest import raises as assert_raises


@testing.with_requires("scipy")
class TestBilinear_zpk:

    @testing.numpy_cupy_allclose(scipy_name="scp")
    def test_basic(self, xp, scp):
        z = [-2j, +2j]
        p = [-0.75, -0.5-0.5j, -0.5+0.5j]
        k = 3

        z_d, p_d, k_d = scp.signal.bilinear_zpk(z, p, k, 10)
        return z_d, p_d, k_d

        """
        assert_allclose(sort(z_d), sort([(20-2j)/(20+2j), (20+2j)/(20-2j),
                                         -1]))
        assert_allclose(sort(p_d), sort([77/83,
                                         (1j/2 + 39/2) / (41/2 - 1j/2),
                                         (39/2 - 1j/2) / (1j/2 + 41/2), ]))
        assert_allclose(k_d, 9696/69803)
        """


@testing.with_requires("scipy")
class TestBilinear:

    @testing.numpy_cupy_allclose(scipy_name="scp")
    def test_basic(self, xp, scp):
        b = [0.14879732743343033]
        a = [1, 0.54552236880522209, 0.14879732743343033]
        b_z, a_z = scp.signal.bilinear(b, a, 0.5)
        return b_z, a_z

#        assert_array_almost_equal(b_z, [0.087821, 0.17564, 0.087821],
#                                  decimal=5)
#        assert_array_almost_equal(a_z, [1, -1.0048, 0.35606], decimal=4)

    @testing.numpy_cupy_allclose(scipy_name="scp")
    def test_basic_2(self, xp, scp):
        b = [1, 0, 0.17407467530697837]
        a = [1, 0.18460575326152251, 0.17407467530697837]
        b_z, a_z = scp.signal.bilinear(b, a, 0.5)
        return b_z, a_z

#        assert_array_almost_equal(b_z, [0.86413, -1.2158, 0.86413],
#                                  decimal=4)
#        assert_array_almost_equal(a_z, [1, -1.2158, 0.72826],
#                                  decimal=4)


@testing.with_requires("scipy")
class TestNormalize:

    def test_allclose(self):
        """Test for false positive on allclose in normalize() in
        filter_design.py"""
        # Test to make sure the allclose call within signal.normalize does not
        # choose false positives. Then check against a known output from MATLAB
        # to make sure the fix doesn't break anything.

        # These are the coefficients returned from
        #   `[b,a] = cheby1(8, 0.5, 0.048)'
        # in MATLAB. There are at least 15 significant figures in each
        # coefficient, so it makes sense to test for errors on the order of
        # 1e-13 (this can always be relaxed if different platforms have
        # different rounding errors)
        b_matlab = cupy.array([2.150733144728282e-11, 1.720586515782626e-10,
                               6.022052805239190e-10, 1.204410561047838e-09,
                               1.505513201309798e-09, 1.204410561047838e-09,
                               6.022052805239190e-10, 1.720586515782626e-10,
                               2.150733144728282e-11])
        a_matlab = cupy.array([1.000000000000000e+00, -7.782402035027959e+00,
                               2.654354569747454e+01, -5.182182531666387e+01,
                               6.334127355102684e+01, -4.963358186631157e+01,
                               2.434862182949389e+01, -6.836925348604676e+00,
                               8.412934944449140e-01])

        # This is the input to signal.normalize after passing through the
        # equivalent steps in signal.iirfilter as was done for MATLAB
        b_norm_in = cupy.array(
            [1.5543135865293012e-06, 1.2434508692234413e-05,
             4.3520780422820447e-05, 8.7041560845640893e-05,
             1.0880195105705122e-04, 8.7041560845640975e-05,
             4.3520780422820447e-05, 1.2434508692234413e-05,
             1.5543135865293012e-06])
        a_norm_in = cupy.array(
            [7.2269025909127173e+04, -5.6242661430467968e+05,
             1.9182761917308895e+06, -3.7451128364682454e+06,
             4.5776121393762771e+06, -3.5869706138592605e+06,
             1.7596511818472347e+06, -4.9409793515707983e+05,
             6.0799461347219651e+04])

        b_output, a_output = signal.normalize(b_norm_in, a_norm_in)

        # The test on b works for decimal=14 but the one for a does not. For
        # the sake of consistency, both of these are decimal=13. If something
        # breaks on another platform, it is probably fine to relax this lower.
        assert_array_almost_equal(b_matlab, b_output, decimal=13)
        assert_array_almost_equal(a_matlab, a_output, decimal=13)

    def test_errors(self):
        """Test the error cases."""
        # all zero denominator
        assert_raises(ValueError, signal.normalize, [1, 2], 0)

        # denominator not 1 dimensional
        assert_raises(ValueError, signal.normalize, [1, 2], [[1]])

        # numerator too many dimensions
        assert_raises(ValueError, signal.normalize, [[[1, 2]]], 1)


@testing.with_requires("scipy")
class TestLp2lp:

    @testing.numpy_cupy_allclose(scipy_name="scp")
    def test_basic(self, xp, scp):
        b = [1]
        a = [1, float(xp.sqrt(2)), 1]
        b_lp, a_lp = scp.signal.lp2lp(b, a, 0.38574256627112119)
        return b_lp, a_lp


@testing.with_requires("scipy")
class TestLp2hp:

    @testing.numpy_cupy_allclose(scipy_name="scp")
    def test_basic(self, xp, scp):
        b = [0.25059432325190018]
        a = [1, 0.59724041654134863, 0.92834805757524175, 0.25059432325190018]
        b_hp, a_hp = scp.signal.lp2hp(b, a, 2*pi*5000)
        return b_hp, a_hp


@testing.with_requires("scipy")
class TestLp2bp:

    @testing.numpy_cupy_allclose(scipy_name="scp")
    def test_basic(self, xp, scp):
        b = [1]
        a = [1, 2, 2, 1]
        b_bp, a_bp = scp.signal.lp2bp(b, a, 2*pi*4000, 2*pi*2000)
        return b_bp, a_bp


@testing.with_requires("scipy")
class TestLp2bs:

    @testing.numpy_cupy_allclose(scipy_name="scp")
    def test_basic(self, xp, scp):
        b = [1]
        a = [1, 1]
        b_bs, a_bs = scp.signal.lp2bs(
            b, a, 0.41722257286366754, 0.18460575326152251)
        return b_bs, a_bs


@testing.with_requires("scipy")
class TestLp2lp_zpk:

    @testing.numpy_cupy_allclose(scipy_name="scp")
    def test_basic(self, xp, scp):
        z = []
        p = [(-1+1j) / sqrt(2), (-1-1j) / sqrt(2)]
        k = 1
        z_lp, p_lp, k_lp = scp.signal.lp2lp_zpk(z, p, k, 5)
        return z_lp, p_lp, k_lp

    @testing.numpy_cupy_allclose(scipy_name="scp")
    def test_basic_2(self, xp, scp):

        # Pseudo-Chebyshev with both poles and zeros
        z = [-2j, +2j]
        p = [-0.75, -0.5-0.5j, -0.5+0.5j]
        k = 3
        z_lp, p_lp, k_lp = scp.signal.lp2lp_zpk(z, p, k, 20)
        return z_lp, p_lp, k_lp


@testing.with_requires("scipy")
class TestLp2hp_zpk:

    @testing.numpy_cupy_allclose(scipy_name="scp")
    def test_basic(self, xp, scp):
        z = []
        p = [(-1+1j)/np.sqrt(2), (-1-1j)/np.sqrt(2)]
        k = 1

        z_hp, p_hp, k_hp = scp.signal.lp2hp_zpk(z, p, k, 5)
        return z_hp, p_hp, k_hp

    @testing.numpy_cupy_allclose(scipy_name="scp")
    def test_basic_2(self, xp, scp):
        z = [-2j, +2j]
        p = [-0.75, -0.5-0.5j, -0.5+0.5j]
        k = 3
        z_hp, p_hp, k_hp = scp.signal.lp2hp_zpk(z, p, k, 6)
        return z_hp, p_hp, k_hp


@testing.with_requires("scipy")
class TestLp2bp_zpk:

    @testing.numpy_cupy_allclose(scipy_name="scp")
    def test_basic(self, xp, scp):
        z = [-2j, +2j]
        p = [-0.75, -0.5-0.5j, -0.5+0.5j]
        k = 3
        z_bp, p_bp, k_bp = scp.signal.lp2bp_zpk(z, p, k, 15, 8)
        return z_bp, p_bp, k_bp


@testing.with_requires("scipy")
class TestLp2bs_zpk:

    @testing.numpy_cupy_allclose(scipy_name="scp")
    def test_basic(self, xp, scp):
        z = [-2j, +2j]
        p = [-0.75, -0.5-0.5j, -0.5+0.5j]
        k = 3

        z_bs, p_bs, k_bs = scp.signal.lp2bs_zpk(z, p, k, 35, 12)

        # complex sort gets confused by real parts of order +/- epsilon
        z_bs_s = z_bs[xp.argsort(z_bs.imag)]
        p_bs_s = p_bs[xp.argsort(p_bs.imag)]
        return z_bs_s, p_bs_s, k_bs


@testing.with_requires("scipy >= 1.8.0")
class TestZpk2Sos:

    @pytest.mark.parametrize('dt', 'fdFD')
    @pytest.mark.parametrize('pairing, analog',
                             [('nearest', False),
                              ('keep_odd', False),
                              ('minimal', False),
                              ('minimal', True)])
    @testing.numpy_cupy_allclose(scipy_name='scp')
    def test_dtypes(self, dt, pairing, analog, xp, scp):
        z = xp.array([-1, -1]).astype(dt)
        ct = dt.upper()  # the poles have to be complex
        p = xp.array([0.57149 + 0.29360j, 0.57149 - 0.29360j]).astype(ct)
        k = xp.array(1).astype(dt)
        sos = scp.signal.zpk2sos(z, p, k, pairing=pairing, analog=analog)
        return sos

    @pytest.mark.parametrize('case', [
        # (z, p, k)
        # cases that match octave
        ([-1, -1], [0.57149 + 0.29360j, 0.57149 - 0.29360j], 1),
        ([1j, -1j], [0.9, -0.9, 0.7j, -0.7j], 1),
        ([], [0.8, -0.5+0.25j, -0.5-0.25j], 1),
        ([1., 1., 0.9j, -0.9j],
         [0.99+0.01j, 0.99-0.01j, 0.1+0.9j, 0.1-0.9j], 1),
        ([0.9+0.1j, 0.9-0.1j, -0.9], [0.75+0.25j, 0.75-0.25j, 0.9], 1),
        # Cases that differ from octave:
        ([-0.3090 + 0.9511j, -0.3090 - 0.9511j, 0.8090 + 0.5878j,
          +0.8090 - 0.5878j, -1.0000 + 0.0000j],
         [-0.3026 + 0.9312j, -0.3026 - 0.9312j, 0.7922 + 0.5755j,
          +0.7922 - 0.5755j, -0.9791 + 0.0000j],
         1),
        ([-1 - 1.4142j, -1 + 1.4142j, -0.625 - 1.0533j, -0.625 + 1.0533j],
         [-0.2 - 0.6782j, -0.2 + 0.6782j, -0.1 - 0.5385j, -0.1 + 0.5385j],
         4),
        ([], [0.2, -0.5+0.25j, -0.5-0.25j], 1.),

    ])
    @pytest.mark.parametrize('pairing', ['nearest', 'keep_odd'])
    @testing.numpy_cupy_allclose(scipy_name='scp', atol=1e-12)
    def test_basic(self, case, pairing, xp, scp):
        z, p, k = case
        z = xp.asarray(z)
        p = xp.asarray(p)
        sos = scp.signal.zpk2sos(z, p, k, pairing=pairing)
        return sos

    @testing.numpy_cupy_allclose(scipy_name='scp', atol=1e-12)
    def test_basic_2(self, xp, scp):
        # The next two examples are adapted from Leland B. Jackson,
        # "Digital Filters and Signal Processing (1995) p.400:
        deg2rad = np.pi / 180.
        k = 1.

        # first example
        thetas = [22.5, 45, 77.5]
        mags = [0.8, 0.6, 0.9]
        z = xp.array([xp.exp(theta * deg2rad * 1j) for theta in thetas])
        z = xp.concatenate((z, z.conj()))
        p = xp.array([mag * xp.exp(theta * deg2rad * 1j)
                      for theta, mag in zip(thetas, mags)])
        p = xp.concatenate((p, p.conj()))
        sos_1 = scp.signal.zpk2sos(z, p, k)

        # second example
        z = xp.array([xp.exp(theta * deg2rad * 1j)
                      for theta in (85., 10.)])
        z = xp.concatenate((z, z.conj(), xp.array([1, -1])))
        sos_2 = scp.signal.zpk2sos(z, p, k)

        return sos_1, sos_2

    @pytest.mark.parametrize('pairing', ['nearest', 'keep_odd', 'minimal'])
    @testing.numpy_cupy_allclose(scipy_name='scp', atol=1e-12)
    def test_pairing(self, pairing, xp, scp):
        # these examples are taken from the docstring, and show the
        # effect of the 'pairing' argument
        z1 = xp.array([-1, -0.5-0.5j, -0.5+0.5j])
        p1 = xp.array([0.75, 0.8+0.1j, 0.8-0.1j])
        sos2 = scp.signal.zpk2sos(z1, p1, 1, pairing=pairing)
        return sos2

    @pytest.mark.parametrize('p',
                             [[-1, 1, -0.1, 0.1],
                              [-0.7071+0.7071j, -0.7071-0.7071j, -0.1j, 0.1j],
                              ])
    @testing.numpy_cupy_allclose(scipy_name='scp', atol=1e-12)
    def test_analog(self, p, xp, scp):
        # test `analog` argument
        # for discrete time, poles closest to unit circle should appear last
        # for cont. time, poles closest to imaginary axis should appear last
        p = xp.asarray(p)
        sos2_dt = scp.signal.zpk2sos([], p, 1, pairing='minimal', analog=False)
        sos2_ct = scp.signal.zpk2sos([], p, 1, pairing='minimal', analog=True)
        return sos2_dt, sos2_ct

    def test_bad_args(self):
        with pytest.raises(ValueError, match=r'pairing must be one of'):
            signal.zpk2sos(cupy.array([1]), cupy.array(
                [2]), 1, pairing='no_such_pairing')

        with pytest.raises(ValueError, match=r'.*pairing must be "minimal"'):
            signal.zpk2sos(cupy.array([1]), cupy.array(
                [2]), 1, pairing='keep_odd', analog=True)

        with pytest.raises(ValueError,
                           match=r'.*must have len\(p\)>=len\(z\)'):
            signal.zpk2sos(cupy.array([1, 1]), cupy.array([2]), 1, analog=True)

        with pytest.raises(ValueError, match=r'k must be real'):
            signal.zpk2sos(cupy.array([1]), cupy.array([2]), k=1j)


@testing.with_requires("scipy")
class TestTf2zpk:

    @pytest.mark.parametrize('dt', ('float64', 'complex128'))
    @testing.numpy_cupy_allclose(scipy_name='scp')
    def test_simple(self, xp, scp, dt):
        z_r = xp.array([0.5, -0.5])
        p_r = xp.array([1.j / sqrt(2), -1.j / sqrt(2)])
        b = xp.poly(z_r).astype(dt)
        a = xp.poly(p_r).real.astype(dt)

        z, p, k = scp.signal.tf2zpk(b, a)
        z.sort()
        # The real part of `p` is ~0.0, so sort by imaginary part
        p = p[xp.argsort(p.imag)]
        return z, p, k


@testing.with_requires("scipy")
class TestSS2TF:

    @testing.numpy_cupy_allclose(scipy_name='scp')
    def test_basic(self, xp, scp):
        # Test a round trip through tf2ss and ss2tf.
        b = xp.array([1.0, 3.0, 5.0])
        a = xp.array([1.0, 2.0, 3.0])

        A, B, C, D = scp.signal.tf2ss(b, a)
        return A, B, C, D

    @testing.numpy_cupy_allclose(scipy_name='scp')
    def test_basic_2(self, xp, scp):
        b = xp.array([1.0, 3.0, 5.0])
        a = xp.array([1.0, 2.0, 3.0])
        A, B, C, D = scp.signal.tf2ss(b, a)

        bb, aa = scp.signal.ss2tf(A, B, C, D)
        return bb, aa

    @testing.numpy_cupy_allclose(scipy_name='scp')
    def test_zero_order_round_trip(self, xp, scp):
        # See gh-5760
        tf = (2, 1)
        A, B, C, D = scp.signal.tf2ss(*tf)
        return A, B, C, D

    @testing.numpy_cupy_allclose(scipy_name='scp')
    def test_zero_order_round_trip_2(self, xp, scp):
        tf = (2, 1)
        A, B, C, D = scp.signal.tf2ss(*tf)

        num, den = scp.signal.ss2tf(A, B, C, D)
        return num, den

    @testing.numpy_cupy_allclose(scipy_name='scp')
    def test_zero_order_round_trip_3(self, xp, scp):
        tf = (xp.asarray([[5], [2]]), 1)
        A, B, C, D = scp.signal.tf2ss(*tf)
        return A, B, C, D

    @testing.numpy_cupy_allclose(scipy_name='scp')
    def test_zero_order_round_trip_4(self, xp, scp):
        tf = (xp.asarray([[5], [2]]), 1)
        A, B, C, D = scp.signal.tf2ss(*tf)

        num, den = scp.signal.ss2tf(A, B, C, D)
        return num, den

    @pytest.mark.parametrize('tf',
                             [([[1, 2], [1, 1]], [1, 2]),
                              ([[1, 0, 1], [1, 1, 1]], [1, 1, 1]),
                                 ([[1, 2, 3], [1, 2, 3]], [1, 2, 3, 4]),
                              ])
    @testing.numpy_cupy_allclose(scipy_name='scp')
    def test_simo_round_trip(self, xp, scp, tf):
        # See gh-5753
        tf = tuple(xp.asarray(x) for x in tf)
        A, B, C, D = scp.signal.tf2ss(*tf)
        return A, B, C, D

    @pytest.mark.parametrize('tf',
                             [([[1, 2], [1, 1]], [1, 2]),
                              ([[1, 0, 1], [1, 1, 1]], [1, 1, 1]),
                                 ([[1, 2, 3], [1, 2, 3]], [1, 2, 3, 4]),
                              ])
    @testing.numpy_cupy_allclose(scipy_name='scp', atol=1e-14)
    def test_simo_round_trip_2(self, xp, scp, tf):
        tf = tuple(xp.asarray(x) for x in tf)
        A, B, C, D = scp.signal.tf2ss(*tf)

        num, den = scp.signal.ss2tf(A, B, C, D)
        return num, den

    @testing.numpy_cupy_allclose(scipy_name='scp', atol=1e-14)
    def test_all_int_arrays(self, xp, scp):
        A = xp.asarray([[0, 1, 0], [0, 0, 1], [-3, -4, -2]])
        B = xp.asarray([[0], [0], [1]])
        C = xp.asarray([[5, 1, 0]])
        D = xp.asarray([[0]])
        num, den = scp.signal.ss2tf(A, B, C, D)
        return num, den

    @testing.numpy_cupy_allclose(scipy_name='scp', atol=1e-14)
    def test_multioutput(self, xp, scp):
        # Regression test for gh-2669.

        # 4 states
        A = xp.array([[-1.0, 0.0, 1.0, 0.0],
                      [-1.0, 0.0, 2.0, 0.0],
                      [-4.0, 0.0, 3.0, 0.0],
                      [-8.0, 8.0, 0.0, 4.0]])

        # 1 input
        B = xp.array([[0.3],
                      [0.0],
                      [7.0],
                      [0.0]])

        # 3 outputs
        C = xp.array([[0.0, 1.0, 0.0, 0.0],
                      [0.0, 0.0, 0.0, 1.0],
                      [8.0, 8.0, 0.0, 0.0]])

        D = xp.array([[0.0],
                      [0.0],
                      [1.0]])

        # Get the transfer functions for all the outputs in one call.
        b_all, a = scp.signal.ss2tf(A, B, C, D)
        return b_all, a


@testing.with_requires('scipy')
class TestSos2Zpk:

    @testing.numpy_cupy_allclose(scipy_name='scp')
    def test_basic(self, xp, scp):
        sos = xp.asarray([[1, 0, 1, 1, 0, -0.81],
                          [1, 0, 0, 1, 0, +0.49]])
        z, p, k = scp.signal.sos2zpk(sos)
        return z, p, k

    @testing.numpy_cupy_allclose(scipy_name='scp')
    def test_basic_2(self, xp, scp):
        sos = [[1.00000, +0.61803, 1.0000, 1.00000, +0.60515, 0.95873],
               [1.00000, -1.61803, 1.0000, 1.00000, -1.58430, 0.95873],
               [1.00000, +1.00000, 0.0000, 1.00000, +0.97915, 0.00000]]
        sos = xp.asarray(sos)
        z, p, k = scp.signal.sos2zpk(sos)
        return z, p, k

    @testing.numpy_cupy_allclose(scipy_name='scp')
    def test_basic_3(self, xp, scp):
        sos = xp.array([[1, 2, 3, 1, 0.2, 0.3],
                        [4, 5, 6, 1, 0.4, 0.5]])
        z2, p2, k2 = scp.signal.sos2zpk(sos)
        return z2, p2, k2

    @testing.numpy_cupy_allclose(scipy_name='scp', atol=1e-14)
    def test_fewer_zeros(self, xp, scp):
        """Test not the expected number of p/z (effectively at origin)."""
        sos = scp.signal.butter(3, 0.1, output='sos')
        z, p, k = scp.signal.sos2zpk(sos)
        return z, p, k

    def test_fewer_zeros_2(self):
        sos = signal.butter(12, [5., 30.], 'bandpass', fs=1200., analog=False,
                            output='sos')
        with pytest.warns(signal.BadCoefficients, match='Badly conditioned'):
            z, p, k = signal.sos2zpk(sos)

        assert len(z) == 24
        assert len(p) == 24


@testing.with_requires("scipy")
class TestSos2Tf:

    @testing.numpy_cupy_allclose(scipy_name='scp')
    def test_basic(self, xp, scp):
        sos = xp.array([[1, 1, 1, 1, 0, -1],
                        [-2, 3, 1, 1, 10, 1]])
        b, a = scp.signal.sos2tf(sos)
        return b, a


@testing.with_requires("scipy")
class TestCplxReal:
    # _cplxreal is a private function, vendored from
    # scipy.signal._filter_design. This test class is also vendored.
    def test_trivial_input(self):
        assert all(x.size == 0 for x in _cplxreal([]))

        cplx1 = _cplxreal(1)
        assert cplx1[0].size == 0
        testing.assert_allclose(cplx1[1], cupy.array([1]))

    def test_output_order(self):
        # zc, zr = _cplxreal(np.roots(array([1, 0, 0, 1])))
        # assert_allclose(np.append(zc, zr), [1/2 + 1j*sin(pi/3), -1])

        eps = cupy.finfo(float).eps  # spacing(1)

        a = [0+1j, 0-1j, eps + 1j, eps - 1j, -eps + 1j, -eps - 1j,
             1, 4, 2, 3, 0, 0,
             2+3j, 2-3j,
             1-eps + 1j, 1+2j, 1-2j, 1+eps - 1j,  # sorts out of order
             3+1j, 3+1j, 3+1j, 3-1j, 3-1j, 3-1j,
             2-3j, 2+3j]
        a = cupy.array(a)
        zc, zr = _cplxreal(a)
        testing.assert_allclose(zc, [1j, 1j, 1j, 1+1j, 1+2j, 2+3j, 2+3j, 3+1j,
                                     3+1j, 3+1j])
        testing.assert_allclose(zr, [0, 0, 1, 2, 3, 4])

        z = cupy.array([1-eps + 1j, 1+2j, 1-2j, 1+eps - 1j, 1+eps+3j,
                        1-2*eps-3j, 0+1j, 0-1j, 2+4j, 2-4j, 2+3j, 2-3j, 3+7j,
                        3-7j, 4-eps+1j, 4+eps-2j, 4-1j, 4-eps+2j])

        zc, zr = _cplxreal(z)
        testing.assert_allclose(zc, [1j, 1+1j, 1+2j, 1+3j, 2+3j, 2+4j, 3+7j,
                                     4+1j, 4+2j])
        assert zr.size == 0

    def test_unmatched_conjugates(self):
        # 1+2j is unmatched
        assert_raises(ValueError, _cplxreal, [1+3j, 1-3j, 1+2j])

        # 1+2j and 1-3j are unmatched
        assert_raises(ValueError, _cplxreal, [1+3j, 1-3j, 1+2j, 1-3j])

        # 1+3j is unmatched
        assert_raises(ValueError, _cplxreal, [1+3j, 1-3j, 1+3j])

        # No pairs
        assert_raises(ValueError, _cplxreal, [1+3j])
        assert_raises(ValueError, _cplxreal, [1-3j])

    def test_real_integer_input(self):
        zc, zr = _cplxreal([2, 0, 1, 4])
        assert zc.size == 0
        testing.assert_allclose(zr, [0, 1, 2, 4], atol=1e-15)


@testing.with_requires("scipy")
class TestLowLevelAP:
    @testing.numpy_cupy_allclose(scipy_name="scp")
    def test_buttap(self, xp, scp):
        return scp.signal.buttap(3)

    @testing.numpy_cupy_allclose(scipy_name="scp")
    def test_cheb1ap(self, xp, scp):
        return scp.signal.cheb1ap(3, 1)

    @testing.numpy_cupy_allclose(scipy_name="scp")
    def test_cheb2ap(self, xp, scp):
        return scp.signal.cheb2ap(3, 1)

    @testing.numpy_cupy_allclose(scipy_name="scp", atol=2e-4, rtol=2e-4)
    def test_ellipap(self, xp, scp):
        return scp.signal.ellipap(7, 1, 10)



from cupy._core.core cimport _ndarray_base


cdef _ndarray_base _ndarray_conj(_ndarray_base self)
cdef _ndarray_base _ndarray_real_getter(_ndarray_base self)
cdef _ndarray_base _ndarray_real_setter(_ndarray_base self, value)
cdef _ndarray_base _ndarray_imag_getter(_ndarray_base self)
cdef _ndarray_base _ndarray_imag_setter(_ndarray_base self, value)
cdef _ndarray_base _ndarray_prod(
    _ndarray_base self, axis, dtype, out, keepdims)
cdef _ndarray_base _ndarray_sum(_ndarray_base self, axis, dtype, out, keepdims)
cdef _ndarray_base _ndarray_cumsum(_ndarray_base self, axis, dtype, out)
cdef _ndarray_base _ndarray_cumprod(_ndarray_base self, axis, dtype, out)
cdef _ndarray_base _ndarray_clip(_ndarray_base self, a_min, a_max, out)

cpdef _ndarray_base _nansum(_ndarray_base a, axis, dtype, out, keepdims)
cpdef _ndarray_base _nanprod(_ndarray_base a, axis, dtype, out, keepdims)

cpdef enum scan_op:
    SCAN_SUM = 0
    SCAN_PROD = 1

cdef _ndarray_base scan(_ndarray_base a, op, dtype=*, _ndarray_base out=*,
                        incomplete=*, chunk_size=*)
cdef object _sum_auto_dtype
cdef object _add
cdef object _conj
cdef object _angle
cdef object _positive
cdef object _negative
cdef object _multiply
cdef object _divide
cdef object _power
cdef object _subtract
cdef object _true_divide
cdef object _floor_divide
cdef object _remainder
cdef object _absolute
cdef object _sqrt



import numpy
import pytest

from cupy import testing
from cupy.cuda import driver
from cupy.cuda import runtime

import cupyx.scipy.signal  # NOQA

try:
    import scipy.signal  # NOQA
except ImportError:
    pass


@testing.with_requires('scipy>=1.9')
class TestSymIIROrder:
    @pytest.mark.parametrize('size', [11, 20, 32, 51, 64, 120])
    @pytest.mark.parametrize(
        'precision', [-1, 2, 1.5, 1.0, 0.5, 0.25, 0.1, 2e-3, 1e-3])
    @testing.for_all_dtypes(no_float16=True, no_bool=True)
    @testing.numpy_cupy_allclose(
        atol=1e-5, rtol=1e-5, scipy_name='scp', accept_error=True)
    def test_symiirorder1(self, size, precision, dtype, xp, scp):
        if xp.dtype(dtype).kind in {'i', 'u'}:
            pytest.skip()
        if (
            runtime.is_hip and driver.get_build_version() < 5_00_00000
        ):
            # ROCm 4.3 raises in Module.get_function()
            pytest.skip()

        x = testing.shaped_random((size,), xp, dtype=dtype)
        c0, z1 = 2.0, 0.5
        return scp.signal.symiirorder1(x, c0, z1, precision)

    @pytest.mark.parametrize('size', [11, 20, 32, 51, 64, 120])
    @pytest.mark.parametrize(
        'precision', [-1, 2, 1.5, 1.0, 0.5, 0.25, 0.1, 2e-3, 1e-3])
    @pytest.mark.parametrize(
        'omega', ['zero', 'pi', 'random']
    )
    @testing.for_dtypes('d')  # XXX: np2.0: f32/f64 dtypes differ
    @testing.numpy_cupy_allclose(
        atol=2e-5, rtol=2e-5, scipy_name='scp', accept_error=True)
    def test_symiirorder2(self, size, precision, omega, dtype, xp, scp):
        if (
            runtime.is_hip and driver.get_build_version() < 5_00_00000
        ):
            # ROCm 4.3 raises in Module.get_function()
            pytest.skip()

        if numpy.lib.NumpyVersion(scipy.__version__) >= "1.15.0":
            if omega in ('zero', 'random'):
                # TODO(asi1024): Fix numerical error
                pytest.xfail(reason="XXX: Large numerical error occurs")

        if omega == 'pi':
            omega = xp.asarray(xp.pi, dtype=dtype)[0]
        elif omega == 'random':
            omega = testing.shaped_random(
                (1,), xp, dtype=dtype, scale=2 * xp.pi)[0]
        else:
            omega = xp.zeros(1, dtype=dtype)[0]

        r = testing.shaped_random((1,), xp, scale=1)[0]
        x = testing.shaped_random((size,), xp, dtype=dtype)

        return scp.signal.symiirorder2(x, r, omega, precision)



import unittest
import pytest

import numpy
import cupy
try:
    import scipy.spatial  # NOQA
    import scipy.spatial.distance  # NOQA
    scipy_available = True
except ModuleNotFoundError:
    scipy_available = False
import cupyx.scipy.spatial.distance  # NOQA
try:
    import cuvs  # NOQA
    cuvs_available = True
except ModuleNotFoundError:
    cuvs_available = False
from cupy import testing


@testing.with_requires("scipy")
@testing.parameterize(*testing.product({
    'dtype': ['float32', 'float64'],
    'rows': [20, 100],
    'cols': [20, 100],
    'metric': ['euclidean', 'cityblock', 'canberra', 'chebyshev',
               'hamming', 'correlation', 'jensenshannon', 'russellrao',
               "minkowski", "cosine", "sqeuclidean"],
    'p': [2.0],
    'order': ["C", "F"]
}))
@pytest.mark.skipif(cupy.cuda.runtime.is_hip, reason="tests for CUDA only")
@pytest.mark.skipif(not scipy_available or not cuvs_available,
                    reason='requires scipy and pylibraft')
class TestCdist(unittest.TestCase):

    def _make_matrix(self, xp, dtype, order):
        shape = (self.rows, self.cols)
        return testing.shaped_random(shape, xp, dtype=dtype,
                                     scale=1, order=order)

    @testing.numpy_cupy_array_almost_equal(decimal=4, scipy_name='scp')
    def test_cdist_(self, xp, scp):

        a = self._make_matrix(xp, self.dtype, self.order)
        b = self._make_matrix(xp, self.dtype, self.order)

        # RussellRao expects boolean arrays
        if self.metric == "russellrao":
            a[a < 0.5] = 0
            a[a >= 0.5] = 1
            b[b < 0.5] = 0
            b[b >= 0.5] = 1

        # JensenShannon expects probability arrays
        elif self.metric == "jensenshannon":
            a_n = a
            b_n = b
            if xp == cupy:
                a_n = a_n.get()
                b_n = b_n.get()

            # l1 normalization is different between cupy and numpy
            # so use numpy.
            norm = numpy.sum(a_n, axis=1)
            a = (a_n.T / norm).T
            if xp == cupy:
                a = cupy.asarray(a)

            norm = numpy.sum(b_n, axis=1)
            b = (b_n.T / norm).T
            if xp == cupy:
                b = cupy.asarray(b)

        if self.metric == 'minkowski':
            out = scp.spatial.distance.cdist(a, b, metric=self.metric,
                                             p=self.p).astype(self.dtype)
        else:
            out = scp.spatial.distance.cdist(a, b, metric=self.metric)\
                .astype(self.dtype)

        print(str(out.shape))
        return out


@testing.with_requires("scipy")
@testing.parameterize(*testing.product({
    'dtype': ['float32', 'float64'],
    'rows': [20, 100],
    'cols': [20, 100],
    'p': [1.0, 2.0, 3.0],
    'order': ["C", "F"]
}))
@pytest.mark.skipif(cupy.cuda.runtime.is_hip, reason="tests for CUDA only")
@pytest.mark.skipif(not scipy_available or not cuvs_available,
                    reason='requires scipy and cuvs')
class TestPdist:

    def _make_matrix(self, xp, dtype, order):
        shape = (self.rows, self.cols)
        out = testing.shaped_random(shape, xp, dtype=dtype,
                                    scale=1, order=order)
        return out

    @testing.numpy_cupy_array_almost_equal(decimal=4, scipy_name='scp')
    def test_pdist_(self, xp, scp):

        a = self._make_matrix(xp, self.dtype, self.order)
        out = scp.spatial.distance.pdist(a).astype(self.dtype)
        return out


@testing.with_requires("scipy")
@testing.parameterize(*testing.product({
    'dtype': ['float32', 'float64'],
    'rows': [20, 100],
    'cols': [20, 100],
    'p': [1.0, 2.0, 3.0],
    'order': ["C", "F"]
}))
@pytest.mark.skipif(cupy.cuda.runtime.is_hip, reason="tests for CUDA only")
@pytest.mark.skipif(not scipy_available or not cuvs_available,
                    reason='requires scipy and cuvs')
class TestDistanceMatrix(unittest.TestCase):

    def _make_matrix(self, xp, dtype, order):
        shape = (self.rows, self.cols)
        return testing.shaped_random(shape, xp, dtype=dtype,
                                     scale=1, order=order)

    @testing.numpy_cupy_array_almost_equal(decimal=4, scipy_name='scp')
    def test_distance_matrix_(self, xp, scp):

        a = self._make_matrix(xp, self.dtype, self.order)
        out = scp.spatial.distance_matrix(a, a, p=self.p).astype(self.dtype)
        return out


@testing.with_requires("scipy")
@testing.parameterize(*testing.product({
    'dtype': ['float32', 'float64'],
    'cols': [20, 100],
    'p': [1.0, 2.0, 3.0],
    'order': ["C", "F"]
}))
@pytest.mark.skipif(cupy.cuda.runtime.is_hip, reason="tests for CUDA only")
@pytest.mark.skipif(not scipy_available or not cuvs_available,
                    reason='requires scipy and cuvs')
class TestDistanceFunction(unittest.TestCase):

    def _make_matrix(self, xp, dtype, order):
        shape = (1, self.cols) if xp is cupy else (self.cols,)
        return testing.shaped_random(shape, xp, dtype=dtype,
                                     scale=1, order=order)

    @testing.numpy_cupy_equal(scipy_name='scp')
    def test_minkowski_(self, xp, scp):

        a = self._make_matrix(xp, self.dtype, self.order)
        out = scp.spatial.distance.minkowski(a, a, p=self.p)
        return out

    @testing.numpy_cupy_equal(scipy_name='scp')
    def test_canberra_(self, xp, scp):

        a = self._make_matrix(xp, self.dtype, self.order)
        out = scp.spatial.distance.canberra(a, a)
        return out

    @testing.numpy_cupy_equal(scipy_name='scp')
    def test_chebyshev_(self, xp, scp):

        a = self._make_matrix(xp, self.dtype, self.order)
        out = scp.spatial.distance.chebyshev(a, a)
        return out

    @testing.numpy_cupy_equal(scipy_name='scp')
    def test_cityblock_(self, xp, scp):

        a = self._make_matrix(xp, self.dtype, self.order)
        out = scp.spatial.distance.cityblock(a, a)
        return out

    @testing.numpy_cupy_array_almost_equal(scipy_name='scp', type_check=False)
    def test_correlation_(self, xp, scp):

        a = self._make_matrix(xp, self.dtype, self.order)
        out = scp.spatial.distance.correlation(a, a)
        return xp.asarray(out)

    @testing.numpy_cupy_array_almost_equal(scipy_name='scp', type_check=False)
    def test_cosine_(self, xp, scp):

        a = self._make_matrix(xp, self.dtype, self.order)
        out = scp.spatial.distance.cosine(a, a)
        return xp.asarray(out)

    @testing.numpy_cupy_equal(scipy_name='scp')
    def test_hamming_(self, xp, scp):

        a = self._make_matrix(xp, self.dtype, self.order)
        out = scp.spatial.distance.hamming(a, a)
        return out

    @testing.numpy_cupy_equal(scipy_name='scp')
    def test_euclidean_(self, xp, scp):

        a = self._make_matrix(xp, self.dtype, self.order)
        out = scp.spatial.distance.euclidean(a, a)
        return out

    @testing.numpy_cupy_equal(scipy_name='scp')
    def test_jensenshannon_(self, xp, scp):

        a = self._make_matrix(xp, self.dtype, self.order)
        out = scp.spatial.distance.jensenshannon(a, a)
        return out

    @testing.numpy_cupy_array_almost_equal(scipy_name='scp', type_check=False)
    def test_russellrao_(self, xp, scp):

        a = self._make_matrix(xp, self.dtype, self.order)
        a[a < 0.5] = 0
        a[a >= 0.5] = 1
        out = scp.spatial.distance.russellrao(a, a)
        return xp.asarray(out)

    @testing.numpy_cupy_equal(scipy_name='scp')
    def test_sqeuclidean_(self, xp, scp):

        a = self._make_matrix(xp, self.dtype, self.order)
        out = scp.spatial.distance.sqeuclidean(a, a)
        return out



import unittest

import cupy
from cupy import testing
import cupyx.scipy.linalg

import numpy
import pytest

try:
    import scipy.linalg
    _scipy_available = True
except ImportError:
    _scipy_available = False


@testing.parameterize(*testing.product({
    'trans': [0, 1, 2, 'N', 'T', 'C'],
    'lower': [True, False],
    'unit_diagonal': [True, False],
    'overwrite_b': [True, False],
    'check_finite': [True, False],
}))
@testing.with_requires('scipy')
class TestSolveTriangular(unittest.TestCase):

    @testing.for_dtypes('fdFD')
    def check_x(self, a_shape, b_shape, dtype):
        a_cpu = numpy.random.randint(1, 10, size=a_shape).astype(dtype)
        b_cpu = numpy.random.randint(1, 10, size=b_shape).astype(dtype)
        a_cpu = numpy.tril(a_cpu)

        if self.lower is False:
            a_cpu = a_cpu.T
        if self.unit_diagonal is True:
            numpy.fill_diagonal(a_cpu, 1)

        a_gpu = cupy.asarray(a_cpu)
        b_gpu = cupy.asarray(b_cpu)
        a_gpu_copy = a_gpu.copy()
        b_gpu_copy = b_gpu.copy()
        result_cpu = scipy.linalg.solve_triangular(
            a_cpu, b_cpu, trans=self.trans, lower=self.lower,
            unit_diagonal=self.unit_diagonal, overwrite_b=self.overwrite_b,
            check_finite=self.check_finite)
        result_gpu = cupyx.scipy.linalg.solve_triangular(
            a_gpu, b_gpu, trans=self.trans, lower=self.lower,
            unit_diagonal=self.unit_diagonal, overwrite_b=self.overwrite_b,
            check_finite=self.check_finite)
        assert result_cpu.dtype == result_gpu.dtype
        cupy.testing.assert_allclose(result_cpu, result_gpu, atol=1e-3)
        cupy.testing.assert_array_equal(a_gpu_copy, a_gpu)
        if not self.overwrite_b:
            cupy.testing.assert_array_equal(b_gpu_copy, b_gpu)

    def test_solve(self):
        self.check_x((4, 4), (4,))
        self.check_x((5, 5), (5, 2))
        self.check_x((5, 5), (5, 5))

    @pytest.mark.xfail(strict=True, reason="not implemented")
    def test_solve_batchwise(self):
        self.check_x((3, 3, 3), (3,))
        self.check_x((3, 3, 3), (3, 2))
        self.check_x((8, 3, 3), (3, 2))
        self.check_x((3, 3), (8, 3, 2))
        self.check_x((8, 3, 3), (8, 3, 2))
        self.check_x((2, 1, 3, 3), (1, 4, 3, 2))

    def check_shape(self, a_shape, b_shape):
        for xp, sp in ((numpy, scipy), (cupy, cupyx.scipy)):
            a = xp.random.rand(*a_shape)
            b = xp.random.rand(*b_shape)
            with pytest.raises(ValueError):
                sp.linalg.solve_triangular(
                    a, b, trans=self.trans, lower=self.lower,
                    unit_diagonal=self.unit_diagonal,
                    overwrite_b=self.overwrite_b,
                    check_finite=self.check_finite)

    def test_invalid_shape(self):
        self.check_shape((2, 3), (4,))
        self.check_shape((3, 3), (2,))
        self.check_shape((3, 3), (2, 2))
        self.check_shape((3, 3, 4), (3,))
        self.check_shape((3, 3, 3), (2, 3))

    def check_infinite(self, a_shape, b_shape):
        for xp, sp in ((numpy, scipy), (cupy, cupyx.scipy)):
            a = xp.random.rand(*a_shape)
            b = xp.random.rand(*b_shape)
            a[(0,) * a.ndim] = numpy.inf
            b[(0,) * b.ndim] = numpy.inf
            with pytest.raises(ValueError):
                sp.linalg.solve_triangular(
                    a, b, trans=self.trans, lower=self.lower,
                    unit_diagonal=self.unit_diagonal,
                    overwrite_b=self.overwrite_b,
                    check_finite=self.check_finite)

    def test_infinite(self):
        if self.check_finite:
            self.check_infinite((4, 4), (4,))
            self.check_infinite((5, 5), (5, 2))
            self.check_infinite((5, 5), (5, 5))



import threading
import warnings

import cupy
from cupy import _util
from cupy._core import _accelerator
from cupy.cuda.device import Handle


_tls = threading.local()

cuquantum = cutensornet = tensornet = None


def _maybe_lazy_load_cutensornet():
    global cuquantum, cutensornet, tensornet

    if cuquantum is not None:
        return

    try:
        import cuquantum
        if hasattr(cuquantum, 'bindings'):
            # cuquantum-python >= 25.03
            from cuquantum.bindings import cutensornet  # binding module
            from cuquantum import tensornet  # module for pythonic APIs
        else:
            # for cuquantum < 25.03, bindings & pythonic APIs
            # all reside under cuquantum.cutensornet
            from cuquantum import cutensornet
            tensornet = cutensornet
    except ImportError:
        pass


@_util.memoize()
def _is_cuqnt_22_11_or_higher():
    ver = [int(i) for i in cuquantum.__version__.split('.')]
    if (ver[0] > 22) or (ver[0] == 22 and ver[1] >= 11):
        return True
    return False


def _is_nonblocking_supported():
    return _is_cuqnt_22_11_or_higher()


def _get_einsum_operands(args):
    """Parse & retrieve einsum operands, assuming ``args`` is in either
    "subscript" or "interleaved" format.
    """

    if len(args) == 0:
        raise ValueError(
            'must specify the einstein sum subscripts string and at least one '
            'operand, or at least one operand and its corresponding '
            'subscripts list')

    if isinstance(args[0], str):
        expr = args[0]
        operands = list(args[1:])
        return expr, operands
    else:
        args = list(args)
        operands = []
        inputs = []
        output = None
        while len(args) >= 2:
            operands.append(args.pop(0))
            inputs.append(args.pop(0))
        if len(args) == 1:
            output = args.pop(0)
        assert not args
        return inputs, operands, output


def _try_use_cutensornet(*args, **kwargs):
    if cupy.cuda.runtime.is_hip:
        return None

    if (_accelerator.ACCELERATOR_CUTENSORNET not in
            _accelerator.get_routine_accelerators()):
        return None

    _maybe_lazy_load_cutensornet()

    if cutensornet is None:
        warnings.warn(
            'using the cuTensorNet backend was requested but it cannot be '
            'imported -- maybe you forgot to install cuQuantum Python? '
            'Please do "pip install cuquantum-python" or "conda install '
            '-c conda-forge cuquantum-python" and retry',
            stacklevel=2)
        return None

    # cannot pop as we might still need kwargs later
    dtype = kwargs.get('dtype', None)
    path = kwargs.get('optimize', False)
    if path is True:
        path = 'greedy'

    # we do very lightweight pre-processing here just to inspect the
    # operands; the actual input verification is deferred to cuTensorNet
    # which can generate far better diagonostic messages
    args = _get_einsum_operands(args)
    operands = [cupy.asarray(op) for op in args[1]]

    if len(operands) == 1:
        # As of cuTENSOR 1.5.0 it still chokes with some common operations
        # like trace ("ii->") so it's easier to just skip all single-operand
        # cases instead of whitelisting what could be done explicitly
        return None

    if (any(op.size == 0 for op in operands) or
            any(len(op.shape) == 0 for op in operands)):
        # To cuTensorNet the shape is invalid
        return None

    # all input dtypes must be identical (to a numerical dtype)
    result_dtype = cupy.result_type(*operands) if dtype is None else dtype
    if result_dtype not in (
            cupy.float32, cupy.float64, cupy.complex64, cupy.complex128):
        return None
    operands = [op.astype(result_dtype, copy=False) for op in operands]

    # prepare cutn inputs
    device = cupy.cuda.runtime.getDevice()
    if not hasattr(_tls, "cutn_handle_cache"):
        cutn_handle_cache = _tls.cutn_handle_cache = {}
    else:
        cutn_handle_cache = _tls.cutn_handle_cache
    handle = cutn_handle_cache.get(device)
    if handle is None:
        handle = cutensornet.create()
        cutn_handle_cache[device] = Handle(handle, cutensornet.destroy)
    else:
        handle = handle.handle
    cutn_options = {'device_id': device, 'handle': handle}
    if _is_nonblocking_supported():
        cutn_options['blocking'] = "auto"

    # TODO(leofang): support all valid combinations:
    # - path from user, contract with cutn (done)
    # - path from cupy, contract with cutn (not yet)
    # - path from cutn, contract with cutn (done)
    # - path from cutn, contract with cupy (not yet)
    raise_warning = False
    if path is False:
        # following the same convention (contracting from the right) as would
        # be produced by _iter_path_pairs(), but converting to a list of pairs
        # due to cuTensorNet's requirement
        path = [(i-1, i-2) for i in range(len(operands), 1, -1)]
    elif len(path) and path[0] == 'einsum_path':
        # let cuTensorNet check if the format is correct
        path = path[1:]
    elif len(path) == 2:
        if isinstance(path[1], (int, float)):
            raise_warning = True
        if path[0] != 'cutensornet':
            raise_warning = True
        path = None
    else:  # path is a string
        if path != 'cutensornet':
            raise_warning = True
        path = None
    if raise_warning:
        warnings.warn(
            'the cuTensorNet backend ignores the "optimize" option '
            'except when an explicit contraction path is provided '
            'or when optimize=False (disable optimization); also, '
            'the maximum intermediate size, if set, is ignored',
            stacklevel=2)
    cutn_optimizer = {'path': path} if path else None

    if len(args) == 2:
        out = tensornet.contract(
            args[0], *operands, options=cutn_options, optimize=cutn_optimizer)
    elif len(args) == 3:
        inputs = [i for pair in zip(operands, args[0]) for i in pair]
        if args[2] is not None:
            inputs.append(args[2])
        out = tensornet.contract(
            *inputs, options=cutn_options, optimize=cutn_optimizer)
    else:
        assert False

    return out
